import streamlit as st
import plotly.graph_objects as go

st.title(":material/blur_on: Neutron corrections")

if not st.session_state["data_parsed"]:
    st.warning("You need to parse the data files first.")
else:

    ##############################
    st.subheader("1. :material/trending_down: Incoming cosmic-ray reference")
    ##############################

    stations = dict(
        JUNG=(46.55, 7.98),
        SOPO=(-90, 0),
        OULU=(65.0544, 25.4681),
        PSNM=(18.59, 98.49),
        MXCO=(19.8, -99.1781),
        HRMS=(-34.43, 19.23),
    )

    @st.fragment
    def select_nm():
        colnm1, colnm2 = st.columns(2, vertical_alignment="top")

        nmdbstation_is = st.session_state[
            "yaml"
        ].process_config.correction_steps.incoming_radiation.reference_neutron_monitor.station
        nmdbstation = colnm1.pills(
            "Select a nearby high-energy neutron monitor",
            options=stations.keys(),
            default="JUNG",
        )
        if nmdbstation != nmdbstation_is:
            st.session_state[
                "yaml"
            ].process_config.correction_steps.incoming_radiation.reference_neutron_monitor.station = (
                nmdbstation
            )

        fig = go.Figure(
            go.Scattergeo(
                lat=[ll[0] for ll in stations.values()],
                lon=[ll[1] for ll in stations.values()],
                marker=dict(color="blue"),
                name="Available stations",
            )
        )
        fig.add_trace(
            go.Scattergeo(
                lat=[stations[nmdbstation][0]],
                lon=[stations[nmdbstation][1]],
                marker=dict(color="orange"),
                name="Selected station",
            )
        )
        fig.add_trace(
            go.Scattergeo(
                lat=[
                    st.session_state["yaml"].sensor_config.sensor_info.latitude
                ],
                lon=[
                    st.session_state[
                        "yaml"
                    ].sensor_config.sensor_info.longitude
                ],
                marker=dict(color="red", symbol="star"),
                name="CRNS location",
            )
        )

        # editing the marker
        fig.update_traces(marker_size=10)

        # this projection_type = 'orthographic is the projection which return 3d globe map'
        fig.update_geos(
            projection=dict(
                type="orthographic",
                rotation=dict(
                    lat=stations[nmdbstation][0], lon=stations[nmdbstation][1]
                ),  # , roll=15),
            )
        )

        # layout, exporting html and showing the plot
        fig.update_layout(
            height=200,
            margin={"r": 0, "t": 0, "l": 0, "b": 0},
            legend=dict(
                yanchor="bottom",
                y=0.0,
            ),
        )

        colnm2.plotly_chart(fig)

    select_nm()

    # @st.cache_data(show_spinner="Downloading from NMDB...")
    def attach_nmdb(station="JUNG"):
        st.session_state["yaml"]._attach_nmdb_data(st.session_state["yaml"].data_hub)

    c1, c2 = st.columns([1, 2])

    if c1.button(":material/download: Attach cosmic-ray data", type="primary"):
        attach_nmdb(
            station=st.session_state[
                "yaml"
            ].process_config.correction_steps.incoming_radiation.reference_neutron_monitor.station
        )

    if (
        "incoming_neutron_intensity"
        in st.session_state["yaml"].data_hub.crns_data_frame.columns
    ):
        st.session_state["data_nmdb_attached"] = True
        import plotly.express as px

        (tab1,) = st.tabs([":material/show_chart: Plots"])

        tab1.plotly_chart(
            px.line(
                st.session_state["yaml"].data_hub.crns_data_frame,
                y="incoming_neutron_intensity",
            ),
            use_container_width=True,
            color_discrete_sequence=["red"],
        )

    with c2.popover(":material/help: Learn more"):
        st.markdown(
            ":material/info: The cosmic-ray reference signal is used to correct the CRNS data for incoming variation of particles, e.g. due to solar events or the solar cycle. This reference signal is measured independently by so-called neutron monitors. The signal of a neutron monitor at a similar geomagnetic cutoff-rigidity and altitude compared to the CRNS location represents the incoming flux at the CRNS site in the best way. Please visit the [NMDB station map](http://www01.nmdb.eu/nest/help.php#helpstations) for more information."
        )
        st.markdown(
            ":material/left_click: Select a nearby neutron monitor, so that its data can be downloaded from [NMDB](http://www01.nmdb.eu/nest) and attached to the CRNS data frame. "
        )

if st.session_state["data_nmdb_attached"]:

    #################################################
    st.subheader("2. :material/flag: Quality checks")
    #################################################

    c1, c2 = st.columns(2)

    # Spike detection period
    c1.number_input(
        label="Spike detection period",
        value=st.session_state[
            "yaml"
        ].process_config.neutron_quality_assessment.raw_neutrons.spike_uni_lof.periods_in_calculation,
        key="input_quality_lof_periods",
        min_value=0,
        max_value=100,
        step=1,
    )

    # Spike detection threshold
    c2.number_input(
        label="Spike detection threshold",
        value=st.session_state[
            "yaml"
        ].process_config.neutron_quality_assessment.raw_neutrons.spike_uni_lof.threshold,
        key="input_quality_lof_threshold",
        min_value=0.0,
        max_value=2.0,
        step=0.01,
    )

    @st.cache_data(show_spinner="Checking quality...")
    def make_quality_check():

        st.session_state["yaml"]._prepare_static_values(st.session_state["yaml"].data_hub)

        from neptoon.quality_control import (
            QualityAssessmentFlagBuilder,
            QualityCheck,
            QATarget,
            QAMethod,
        )

        qa_flags = QualityAssessmentFlagBuilder()
        qa_flags.add_check(
            QualityCheck(
                target=QATarget.RELATIVE_HUMIDITY,
                method=QAMethod.RANGE_CHECK,
                parameters={"min": 0, "max": 100},
            ),
            QualityCheck(
                target=QATarget.RAW_EPI_NEUTRONS,
                method=QAMethod.SPIKE_UNILOF,
                parameters={
                    "periods_in_calculation": st.session_state[
                        "input_quality_lof_periods"
                    ],
                    "threshold": st.session_state[
                        "input_quality_lof_threshold"
                    ],
                },
            ),
        )

        st.session_state["yaml"].data_hub.add_quality_flags(
            custom_flags=qa_flags
        )
        st.session_state["yaml"].data_hub.apply_quality_flags()

        st.session_state[
            "yaml"
        ].process_config.neutron_quality_assessment.raw_neutrons.spike_uni_lof.periods_in_calculation = st.session_state[
            "input_quality_lof_periods"
        ]
        st.session_state[
            "yaml"
        ].process_config.neutron_quality_assessment.raw_neutrons.spike_uni_lof.threshold = st.session_state[
            "input_quality_lof_threshold"
        ]
        st.session_state["data_quality_checked"] = True

    if st.button(":material/flag: Quality check", type="primary"):
        make_quality_check()

if st.session_state["data_quality_checked"]:
    import plotly.express as px

    tab1, tab2, tab3 = st.tabs(
        [
            ":material/Table: Raw data table",
            ":material/flag: Flagged data",
            ":material/show_chart: Plots",
        ]
    )

    tab1.dataframe(st.session_state["yaml"].data_hub.crns_data_frame)
    tab2.dataframe(st.session_state["yaml"].data_hub.flags_data_frame)
    columns_to_plot = [
        "epithermal_neutrons_raw",
        "epithermal_neutrons_cph",
    ]
    tab3.plotly_chart(
        px.line(
            st.session_state["yaml"].data_hub.crns_data_frame[columns_to_plot],
            y=columns_to_plot,
        ),
        use_container_width=True,
    )

    ###############################################################
    st.subheader("3. :material/vertical_align_center: Corrections")
    ###############################################################

    @st.fragment
    def create_correction_input():
        st.write("**3.1 Air pressure correction**")
        c1, c2 = st.columns([1, 1])
        # pressure_method = c1.segmented_control(
        #     "Method", ["Zreda et al. (2012)"], default="Zreda et al. (2012)"
        # )
        with c2.popover(":material/help: Learn more"):
            st.markdown(
                ":material/info: Air pressure represents the mass of air about the sensor. Every meter of air attenuates the cosmic radiations. The factor to correct for this effect is exponential:"
            )
            st.latex(r"C_p = e^{\beta\,(P_0-P_\text{ref})},")
            st.markdown(
                r"where $\beta$ = 0.0076 and $P_\text{ref}$ = 1013 hPa. See [Zreda et al (2012)](https://doi.org/10.5194/hess-16-4079-2012) for details."
            )

        """
        **3.2 Air humidity correction**
        """
        c1, c2 = st.columns([1, 1])

        # humidity_method = c1.segmented_control(
        #     "Method",
        #     ["Rosolem et al. (2013)"],
        #     default="Rosolem et al. (2013)",
        # )
        with c2.popover(":material/help: Learn more"):
            st.markdown(
                ":material/info: Air humidity represents the number of hydrogen atoms in the air above and around the sensor. They attenuate the cosmic radiation from above and the neutron radiation from the sides. The factor to correct for this effect is linear:"
            )
            st.latex(r"C_h = 1+\alpha\,(h-h_\text{ref}),")
            st.markdown(
                r"where $\alpha$ = 0.0054 and $h_\text{ref}$ = 0 g/m³. See [Rosolem et al (2013)](https://doi.org/10.1175/JHM-D-12-0120.1) for details."
            )

        """
        **3.3 Cosmic-ray incoming correction**
        """
        c1, c2 = st.columns([1, 1])
        # incoming_method = c1.segmented_control(
        #     "Method",
        #     ["Zreda et al. (2012)"],
        #     key="other",
        #     default="Zreda et al. (2012)",
        # )
        with c2.popover(":material/help: Learn more"):
            st.markdown(
                ":material/info: Incoming cosmic radiation varies with time and space depending on the solar activity, for instance. The reference signal is measured by neutron monitors and can be used inversely to correct the CRNS neutrons:"
            )
            st.latex(r"C_I = M_\text{ref}/M\,,")
            st.markdown(
                r"where $M$ is neutron neutron monitor data from [NMDB](http://www01.nmdb.eu/nest/) and $M_\text{ref}$ = 159 cps is a normalization factor. See [Zreda et al (2012)](https://doi.org/10.5194/hess-16-4079-2012) for details."
            )

    create_correction_input()

    @st.cache_data(show_spinner="Making corrections...")
    def make_corrections():

        from neptoon.corrections import (
            CorrectionType,
            CorrectionTheory,
        )

        st.session_state["yaml"].data_hub.select_correction(
            correction_type=CorrectionType.INCOMING_INTENSITY,
            correction_theory=CorrectionTheory.HAWDON_2014,
        )
        st.session_state["yaml"].data_hub.select_correction(
            correction_type=CorrectionType.HUMIDITY,
            correction_theory=CorrectionTheory.ROSOLEM_2013,
        )

        st.session_state["yaml"].data_hub.select_correction(
            correction_type=CorrectionType.PRESSURE,
        )

        st.session_state["yaml"].data_hub.correct_neutrons()
        st.session_state["data_corrections_made"] = True

    if st.button(
        ":material/vertical_align_center: Make corrections", type="primary"
    ):
        make_corrections()

if st.session_state["data_corrections_made"]:

    tab3, tab4, tab5 = st.tabs(
        [
            ":material/Table: Processed data table",
            ":material/show_chart: Correction factors",
            ":material/show_chart: Corrected neutrons",
        ]
    )

    tab3.dataframe(st.session_state["yaml"].data_hub.crns_data_frame)

    selected_columns_corr = [
        "atmospheric_pressure_correction",
        "humidity_correction",
        "incoming_neutron_intensity_correction",
    ]

    data_corr_factors = st.session_state["yaml"].data_hub.crns_data_frame[
        selected_columns_corr
    ]

    tab4.plotly_chart(
        px.line(data_corr_factors, y=selected_columns_corr),
        use_container_width=True,
    )

    # st.session_state["yaml"].data_hub.crns_data_frame.loc[
    #     st.session_state["yaml"].data_hub.crns_data_frame["corrected_epithermal_neutrons"] < 300,
    #     "corrected_epithermal_neutrons",
    # ] = np.nan

    # tab5.write(
    #     "For better visibility, displayed neutrons are smoothed across 24 hours."
    # )
    selected_columns_corrn = [
        "epithermal_neutrons_cph",
        "corrected_epithermal_neutrons",
    ]

    data_corr_neutrons = st.session_state["yaml"].data_hub.crns_data_frame[
        selected_columns_corrn
    ]

    tab5.plotly_chart(
        px.line(
            data_corr_neutrons,  # .rolling(24 * 4).mean(),
            y=selected_columns_corrn,
        ),
        use_container_width=True,
    )

    st.subheader(":material/airwave: Smoothing")

    # neutron_range = (
    #     st.session_state["yaml"]
    #     .data_hub.crns_data_frame["corrected_epithermal_neutrons"]
    #     .min(),
    #     st.session_state["yaml"]
    #     .data_hub.crns_data_frame["corrected_epithermal_neutrons"]
    #     .max(),
    # )

    # from neptoon.columns import ColumnInfo

    column_to_smooth = "corrected_epithermal_neutrons"  # str(ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT_FINAL)

    # st.write(column_to_smooth)

    @st.fragment
    def smooth_neutrons():

        smooth = st.slider("Smoothing window in hours", 1, 25, 1)
        column_smoothed = column_to_smooth + f"_rollingmean_{smooth}"

        st.session_state["yaml"].data_hub.smooth_data(
            column_to_smooth=column_to_smooth,
            smooth_method="rolling_mean",
            window=smooth,
        )

        show_columns = [column_to_smooth]
        if (
            column_smoothed
            in st.session_state["yaml"].data_hub.crns_data_frame
        ):
            show_columns.append(column_smoothed)

        st.plotly_chart(
            px.line(
                st.session_state["yaml"].data_hub.crns_data_frame[
                    show_columns
                ],
                y=show_columns,
                # range_y=neutron_range,
            ),
            use_container_width=True,
        )
        # st.write(st.session_state["yaml"].data_hub.crns_data_frame.columns)

    smooth_neutrons()



import pandas as pd
from neptoon.columns.column_information import ColumnInfo
from neptoon.corrections.theory.air_humidity_corrections import (
    calc_absolute_humidity,
    calc_actual_vapour_pressure,
    calc_saturation_vapour_pressure,
)


class AbsoluteHumidityCreator:
    """
    Given a DataFrame with at least:
      - a temperature column (C)
      - a relative humidity column (in %)

    this class will add:
      1) saturation vapour pressure (hPa),
      2) actual vapour pressure,
      3) absolute humidity (g/m3).
    """

    def __init__(
        self,
        data_frame: pd.DataFrame,
        absolute_hum_col_name: str = str(ColumnInfo.Name.ABSOLUTE_HUMIDITY),
        temperature_col_name: str = str(ColumnInfo.Name.AIR_TEMPERATURE),
        sat_vapour_col_name: str = str(
            ColumnInfo.Name.SATURATION_VAPOUR_PRESSURE
        ),
        actual_vapour_pressure_col_name: str = str(
            ColumnInfo.Name.ACTUAL_VAPOUR_PRESSURE
        ),
        relative_hum_col_name: str = str(
            ColumnInfo.Name.AIR_RELATIVE_HUMIDITY
        ),
    ):
        self.data_frame = data_frame
        self.absolute_hum_col_name = absolute_hum_col_name
        self.temperature_col_name = temperature_col_name
        self.sat_vapour_col_name = sat_vapour_col_name
        self.actual_vapour_pressure_col_name = actual_vapour_pressure_col_name
        self.relative_hum_col_name = relative_hum_col_name

        # self._check_required_columns_available()

    def _check_required_columns_available(self):
        """
        Ensures that the required columns are available before processing.

        Raises
        ------
        KeyError
            Error when required column not available
        """
        missing = []
        for col in (self.temperature_col_name, self.relative_hum_col_name):
            if col not in self.data_frame.columns:
                missing.append(col)
        if missing:
            raise KeyError(
                f"DataFrame is missing required column(s): {missing}"
            )

    def check_if_abs_hum_exists(self):
        """
        Check if absolute humidity is already in the data frame

        Returns
        -------
        bool
        """
        if self.absolute_hum_col_name in self.data_frame.columns:
            abs_hum_exists = True
            return abs_hum_exists

    def create_saturation_vapour_pressure_data(self):
        """
        Creates a column with saturation vapour pressure, by applying
        calc_saturation_vapour_pressure to each temperature.
        """
        self.data_frame[self.sat_vapour_col_name] = self.data_frame[
            self.temperature_col_name
        ].apply(calc_saturation_vapour_pressure)

    def create_actual_vapour_pressure_data(self):
        """
        Creates a column with actual vapour pressure.
        """
        self.data_frame[self.actual_vapour_pressure_col_name] = (
            self.data_frame.apply(
                lambda row: calc_actual_vapour_pressure(
                    saturation_vapour_pressure=row[self.sat_vapour_col_name],
                    relative_humidity=row[self.relative_hum_col_name],
                ),
                axis=1,
            )
        )

    def create_absolute_humidity_data(self):
        """
        Creates a column with absolute humidity data.
        """
        self.data_frame[self.absolute_hum_col_name] = self.data_frame.apply(
            lambda row: calc_absolute_humidity(
                vapour_pressure=row[self.actual_vapour_pressure_col_name],
                temperature=row[self.temperature_col_name],
            ),
            axis=1,
        )

    def check_and_return_abs_hum_column(self):
        """
        Creates absolute humidity data whilst also creating saturation
        vapour pressure data and actual vapour pressure data.

        Returns the whole dataframe with new data attached.

        Returns
        -------
        pd.DataFrame
            DataFrame
        """
        abs_hum_exists = self.check_if_abs_hum_exists()  # returns if exists
        if abs_hum_exists:
            return self.data_frame
        else:
            self._check_required_columns_available()
            self.create_saturation_vapour_pressure_data()
            self.create_actual_vapour_pressure_data()
            self.create_absolute_humidity_data()
            return self.data_frame



import streamlit as st
from pathlib import Path
from neptoon_gui_utils import cleanup, save_uploaded_file
import atexit

st.title(":material/full_stacked_bar_chart: Read data")

if not st.session_state["yaml_checked"]:
    st.warning("You need to select a configuration first.")
else:

    ##############################
    st.subheader("1. Data source")
    ##############################

    use_raw_data = st.session_state[
        "yaml"
    ].sensor_config.raw_data_parse_options.parse_raw_data

    selected_data_type = st.segmented_control(
        "Select data type",
        ["Raw data", "Preformatted data"],
        selection_mode="single",
        default="Raw data" if use_raw_data else "Preformatted data",
    )

    if selected_data_type == "Raw data":

        st.session_state["data_raw_file"] = Path(
            st.session_state[
                "yaml"
            ].sensor_config.raw_data_parse_options.data_location
            or ""
        )

        uploaded_file = st.file_uploader(
            "Upload files",
            type={"csv", "zip"},
            key="data_raw_upload",
        )
        if uploaded_file:
            # File upload
            st.session_state["data_raw_upload_name"] = uploaded_file.name

            temp_file_path = save_uploaded_file(uploaded_file)
            atexit.register(cleanup, temp_file_path)

            st.session_state["data_raw_upload_file"] = temp_file_path
            st.session_state["data_raw_file"] = temp_file_path

        if st.session_state["data_raw_file"]:

            if not st.session_state["data_raw_file"].is_file():
                st.error(
                    "File **{:}** does not exist.".format(
                        st.session_state["data_raw_file"]
                    )
                )
                st.warning("No file selected yet. Please upload your data.")
                st.session_state["data_read_ready"] = False
            else:
                # Already uploaded
                st.success(
                    ":material/check: Using **{:}** as raw data.".format(
                        st.session_state["data_raw_file"]
                    )
                )
                st.session_state["data_read_ready"] = True

                # column_names
                st.text_input(
                    label="Column names",
                    value=", ".join(
                        st.session_state[
                            "yaml"
                        ].sensor_config.raw_data_parse_options.column_names
                        or []
                    ),
                    key="input_dataraw_column_names",
                )

                c1, c2 = st.columns(2)

                # Prefix
                c1.text_input(
                    label="Prefix",
                    value=st.session_state[
                        "yaml"
                    ].sensor_config.raw_data_parse_options.prefix,
                    key="input_dataraw_prefix",
                )

                # Suffix
                c2.text_input(
                    label="Suffix",
                    value=st.session_state[
                        "yaml"
                    ].sensor_config.raw_data_parse_options.suffix,
                    key="input_dataraw_suffix",
                )

                # skip_lines
                c1.text_input(
                    label="Skip lines",
                    value=st.session_state[
                        "yaml"
                    ].sensor_config.raw_data_parse_options.skip_lines,
                    key="input_dataraw_skip_lines",
                )

                # separator
                c2.text_input(
                    label="Separator",
                    value=st.session_state[
                        "yaml"
                    ].sensor_config.raw_data_parse_options.separator,
                    key="input_dataraw_separator",
                )

                with st.expander(
                    "More settings will be editable in future versions."
                ):
                    st.write(
                        st.session_state[
                            "yaml"
                        ].sensor_config.raw_data_parse_options,
                    )

    elif selected_data_type == "Preformatted data":

        st.session_state["data_preformatted_file"] = Path(
            st.session_state[
                "yaml"
            ].sensor_config.time_series_data.path_to_data
            or ""
        )

        uploaded_file = st.file_uploader(
            "Upload files",
            type={"csv", "zip"},
            key="data_preformatted_upload",
        )
        if uploaded_file:
            # File upload
            st.session_state["data_preformatted_upload_name"] = (
                uploaded_file.name
            )

            temp_file_path = save_uploaded_file(uploaded_file)
            atexit.register(cleanup, temp_file_path)

            st.session_state["data_preformatted_upload_file"] = temp_file_path
            st.session_state["data_preformatted_file"] = temp_file_path

        if st.session_state["data_preformatted_file"]:

            if not st.session_state["data_preformatted_file"].is_file():
                st.error(
                    "File **{:}** does not exist.".format(
                        st.session_state["data_preformatted_file"]
                    )
                )
                st.warning("No file selected yet. Please upload your data.")
                st.session_state["data_read_ready"] = False
            else:
                # Already uploaded
                st.success(
                    ":material/check: Using **{:}** as preformatted data.".format(
                        st.session_state["data_preformatted_file"]
                    )
                )
                st.session_state["data_read_ready"] = True

                c1, c2 = st.columns(2)

                # input_resolution
                c1.text_input(
                    label="Input resolution",
                    value=st.session_state[
                        "yaml"
                    ].sensor_config.time_series_data.temporal.input_resolution,
                    key="input_datapre_input_resolution",
                )

                # Output resolution
                c2.text_input(
                    label="Output resolution",
                    value=st.session_state[
                        "yaml"
                    ].sensor_config.time_series_data.temporal.output_resolution,
                    key="input_datapre_output_resolution",
                )

                # date_time_columns
                c1.text_input(
                    label="Datetime columns",
                    value=", ".join(
                        st.session_state[
                            "yaml"
                        ].sensor_config.time_series_data.key_column_info.date_time_columns
                    ),
                    key="input_datapre_date_time_columns",
                )

                # date_time_format
                c2.text_input(
                    label="Datetime format",
                    value=st.session_state[
                        "yaml"
                    ].sensor_config.time_series_data.key_column_info.date_time_format,
                    key="input_datapre_date_time_format",
                )

                with st.expander(
                    "More settings will be editable in future versions."
                ):
                    st.write(
                        st.session_state[
                            "yaml"
                        ].sensor_config.time_series_data.key_column_info,
                    )

    ################################
    st.subheader("2. :material/search_insights: Data inspection")
    ################################

    @st.cache_data(show_spinner="Creating data table...")
    def parse_data():
        # import plotly.express as px

        # with st.spinner("Creating data table..."):
        st.session_state["yaml"].data_hub =  st.session_state["yaml"]._create_data_hub(st.session_state["yaml"].sensor_config)
        data_hub = st.session_state["yaml"].data_hub
        st.write(
            "Parsed {:,.0f} lines and {:.0f} columns of data.".format(
                len(data_hub.crns_data_frame),
                len(data_hub.crns_data_frame.columns),
            )
        )
        st.session_state["data_parsed"] = True

    if st.session_state["data_read_ready"]:
        if st.button(
            ":material/read_more: Apply and parse the data!", type="primary"
        ):
            # st.write(st.session_state["yaml"].sensor_config)
            if use_raw_data:
                st.session_state[
                    "yaml"
                ].sensor_config.raw_data_parse_options.column_names = [
                    x.strip()
                    for x in st.session_state[
                        "input_dataraw_column_names"
                    ].split(",")
                ]
                st.write(
                    st.session_state[
                        "yaml"
                    ].sensor_config.raw_data_parse_options.column_names
                )
                st.session_state[
                    "yaml"
                ].sensor_config.raw_data_parse_options.prefix = st.session_state[
                    "input_dataraw_prefix"
                ]
                st.session_state[
                    "yaml"
                ].sensor_config.raw_data_parse_options.suffix = st.session_state[
                    "input_dataraw_suffix"
                ]
                st.session_state[
                    "yaml"
                ].sensor_config.raw_data_parse_options.skip_lines = int(
                    st.session_state["input_dataraw_skip_lines"]
                )
                st.session_state[
                    "yaml"
                ].sensor_config.raw_data_parse_options.separator = st.session_state[
                    "input_dataraw_separator"
                ]
            else:
                st.session_state[
                    "yaml"
                ].sensor_config.time_series_data.key_column_info.date_time_format = st.session_state[
                    "input_datapre_date_time_format"
                ]
                st.session_state[
                    "yaml"
                ].sensor_config.time_series_data.key_column_info.date_time_columns = [
                    x.strip()
                    for x in st.session_state[
                        "input_datapre_date_time_columns"
                    ].split(",")
                ]
                st.session_state[
                    "yaml"
                ].sensor_config.time_series_data.temporal.output_resolution = st.session_state[
                    "input_datapre_output_resolution"
                ]
                st.session_state[
                    "yaml"
                ].sensor_config.time_series_data.temporal.input_resolution = st.session_state[
                    "input_datapre_input_resolution"
                ]
                # st.success("Changes applied :smile:")

            parse_data()
            # st.write(st.session_state["yaml"].sensor_config)

    @st.fragment
    def make_selection_plot():
        import plotly.express as px

        tab1, tab2 = st.tabs(
            [":material/Table: Raw data table", ":material/show_chart: Plots"]
        )

        tab1.dataframe(data_hub.crns_data_frame)

        selected_columns = tab2.multiselect(
            "Which columns would you like to view?",
            options=data_hub.crns_data_frame.columns,
            default="epithermal_neutrons_cph",
        )

        filtered_data = data_hub.crns_data_frame[selected_columns]

        tab2.plotly_chart(
            px.line(filtered_data, y=selected_columns),
            use_container_width=True,
        )

    if st.session_state["data_parsed"]:
        data_hub = st.session_state["yaml"].data_hub

        make_selection_plot()



import pandas as pd
from dataclasses import dataclass
from neptoon.config.configuration_input import SensorInfo
from neptoon.visulisation.figures import (
    make_nmdb_data_figure,
    soil_moisture_coloured_figure,
    soil_moisture_figure_uncertainty,
    uncorr_and_corr_neutrons_figure,
    atmospheric_conditions_figure,
    correction_factors_figure,
)
from neptoon.columns import ColumnInfo
from typing import List, Optional
from enum import Enum
from magazine import Magazine
from pathlib import Path
import tempfile
import atexit
import shutil


class FigureTopic(Enum):
    NMDB = "nmdb"
    NEUTRONS = "neutrons"
    SOIL_MOISTURE = "soil_moisture"
    ATMOSPHERIC = "atmospheric"


@dataclass
class FigureMetadata:
    """
    Key information on each figure created.

    topic : FigureTopic
        The topic the figure is related to
    description : str
        A brief description about the figure
    required_columns: List[str]
        The columns required for this particular figure.
    method : callable
        The method (found in FigureHandler) which will be called to
        produce this figure.
    """

    topic: FigureTopic
    description: str
    required_columns: List[str]
    method: callable


@dataclass
class TempFigure:
    """
    Tracks a temporary figure file

    name : str
        Name of the figure

    """

    name: str
    path: Path
    topic: str


class TempFigureHandler:
    """
    Handles the creation and storage of figures made with the
    FigureHandler by creating a temp directory, recording paths for each
    figure, and removing temp files upon program exit.
    """

    def __init__(
        self,
    ):
        self._temp_dir = Path(tempfile.mkdtemp(prefix="neptoon_figures_"))
        self.figures: List[TempFigure] = []
        atexit.register(self.cleanup)

    def store_figure(self, name: str, topic: FigureTopic):
        """
        Stores metadata of temporary figure storage

        Parameters
        ----------
        name : str
            name of the figure being stored
        topic : FigureTopic
            Topic the figure is part of (e.g., NMDB, or Soil)

        Returns
        -------
        Path
            Path of the temp figure
        """
        temp_path = self._temp_dir / f"{name}.png"
        self.figures.append(
            TempFigure(
                name=name,
                path=temp_path,
                topic=topic,
            )
        )
        return temp_path

    def cleanup(self):
        """Remove temporary directory and files"""
        if self._temp_dir.exists():
            shutil.rmtree(self._temp_dir)

    def get_figures(self, topic: Optional[FigureTopic] = None):
        """
        Collect the figures that are saved

        Parameters
        ----------
        topic : FigureTopic
            Topic the figure is part of (e.g., NMDB, or Soil)

        Returns
        -------
        List[TempFigure]
            A list of TempFigure objects with location of save files.
        """
        if topic is None:
            return self.figures
        return [fig for fig in self.figures if fig.topic == topic]


class FigureHandler:
    """
    Class to manage creating figures when data has been processed using the CRNSDataHub.
    """

    _figure_registry = {
        "nmdb_incoming_radiation": FigureMetadata(
            topic=FigureTopic.NMDB,
            description="NMDB incoming cosmic radiation plot",
            method="_nmdb_incoming_radiation",
            required_columns=[
                str(ColumnInfo.Name.INCOMING_NEUTRON_INTENSITY),
                str(ColumnInfo.Name.REFERENCE_INCOMING_NEUTRON_VALUE),
            ],
        ),
        "correction_factors_figure": FigureMetadata(
            topic=FigureTopic.NEUTRONS,
            description="Correction factors applied to raw neutron counts",
            method="_neutron_correction_factors",
            required_columns=[],  # empty for flexibility
        ),
        "neutron_counts_corr_uncorr": FigureMetadata(
            topic=FigureTopic.NEUTRONS,
            description="Corrected and uncorrected count rates at the sensor",
            method="_uncorr_and_corr_neutrons_figure",
            required_columns=[
                str(ColumnInfo.Name.EPI_NEUTRON_COUNT_FINAL),
                str(ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT_FINAL),
            ],
        ),
        "soil_moisture_uncertainty": FigureMetadata(
            topic=FigureTopic.SOIL_MOISTURE,
            description="Soil moisture time series with uncertainty bounds",
            method="_soil_moisture_uncertainty",
            required_columns=[
                str(ColumnInfo.Name.SOIL_MOISTURE_VOL_FINAL),
                str(ColumnInfo.Name.SOIL_MOISTURE_UNCERTAINTY_VOL_LOWER),
                str(ColumnInfo.Name.SOIL_MOISTURE_UNCERTAINTY_VOL_UPPER),
            ],
        ),
        # "soil_moisture_coloured": FigureMetadata(
        #     topic=FigureTopic.SOIL_MOISTURE,
        #     description="Soil moisture time series with colour filling",
        #     method="_soil_moisture_colour",
        #     required_columns=[str(ColumnInfo.Name.SOIL_MOISTURE_FINAL)],
        # ),
        "atmospheric_variables": FigureMetadata(
            topic=FigureTopic.ATMOSPHERIC,
            description="Atmospheric variables.",
            method="_atmospheric_variables",
            required_columns=[
                str(ColumnInfo.Name.AIR_PRESSURE),
                str(ColumnInfo.Name.AIR_TEMPERATURE),
                str(ColumnInfo.Name.AIR_RELATIVE_HUMIDITY),
                str(ColumnInfo.Name.ABSOLUTE_HUMIDITY),
            ],
        ),
    }

    def __init__(
        self,
        data_frame: pd.DataFrame,
        sensor_info: SensorInfo,
        create_all: bool = False,
        ignore_sections: List = None,
        selected_figures: List[str] = None,
        show_figures: bool = False,
        backend: str = "Agg",
    ):

        self.data_frame = data_frame
        self.sensor_info = sensor_info
        self.temp_handler = TempFigureHandler()
        self.create_all = create_all
        self.ignore_sections = (
            ignore_sections if ignore_sections is not None else []
        )
        self.selected_figures = (
            selected_figures if selected_figures is not None else []
        )
        self.show_figures = show_figures
        self.backend = backend

    def _validate_required_columns(self, metadata: FigureMetadata):
        """
        Validates that the required columns are present in the data
        frame

        Parameters
        ----------
        metadata : FigureMetadata
            An instance of FigureMetadata

        Raises
        ------
        ValueError
            When missing columns necessary for processing
        """

        missing_cols = [
            col
            for col in metadata.required_columns
            if col not in self.data_frame.columns
        ]
        if missing_cols:
            raise ValueError(
                f"DataFrame is missing required columns: {missing_cols}"
            )

    def _resolve_method(self, metadata: FigureMetadata):
        """
        Resolves the method string to an actual callable method.

        """
        method_name = metadata.method
        if not hasattr(self, method_name):
            raise AttributeError(
                f"Method {method_name} not found in {self.__class__.__name__}"
            )
        return getattr(self, method_name)

    @Magazine.reporting_figure(topic="NMDB")
    def _nmdb_incoming_radiation(self):
        """
        Implements nmdb figure 1
        """
        temp_path = self.temp_handler.store_figure(
            name="nmdb_incoming_radiation",
            topic=FigureTopic.NMDB,
        )
        reference_value = self.data_frame[
            str(ColumnInfo.Name.REFERENCE_INCOMING_NEUTRON_VALUE)
        ].iloc[0]

        make_nmdb_data_figure(
            data_frame=self.data_frame,
            # nmdb_station_name=self.sensor_info.name,
            reference_value=reference_value,
            incoming_neutron_col_name=str(
                ColumnInfo.Name.INCOMING_NEUTRON_INTENSITY
            ),
            show=self.show_figures,
            backend=self.backend,
            save_location=temp_path,
        )

    @Magazine.reporting_figure(topic="Neutron Correction")
    def _uncorr_and_corr_neutrons_figure(self):
        """
        Implements uncorrected and corrected neutrons figures
        """
        temp_path = self.temp_handler.store_figure(
            name="neutron_counts_corr_uncorr",
            topic=FigureTopic.NEUTRONS,
        )
        uncorr_and_corr_neutrons_figure(
            data_frame=self.data_frame,
            station_name=self.sensor_info.name,
            raw_neutron_col_name=str(ColumnInfo.Name.EPI_NEUTRON_COUNT_FINAL),
            corr_neutron_col_name=str(
                ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT_FINAL
            ),
            show=self.show_figures,
            backend=self.backend,
            save_location=temp_path,
        )

    @Magazine.reporting_figure(topic="Soil Moisture")
    def _soil_moisture_uncertainty(self):
        """
        Implements soil moisture with uncertainty figure.
        """
        temp_path = self.temp_handler.store_figure(
            name="soil_moisture_uncertainty",
            topic=FigureTopic.SOIL_MOISTURE,
        )

        sm_max = (
            self.data_frame[str(ColumnInfo.Name.SOIL_MOISTURE_VOL_FINAL)].max()
        ) * 1.1
        sm_min = 0
        sm_range = (sm_min, sm_max)

        soil_moisture_figure_uncertainty(
            data_frame=self.data_frame,
            station_name=self.sensor_info.name,
            soil_moisture_col=str(ColumnInfo.Name.SOIL_MOISTURE_VOL_FINAL),
            upper_uncertainty_col=str(
                ColumnInfo.Name.SOIL_MOISTURE_UNCERTAINTY_VOL_UPPER
            ),
            lower_uncertainty_col=str(
                ColumnInfo.Name.SOIL_MOISTURE_UNCERTAINTY_VOL_LOWER
            ),
            sm_range=sm_range,
            show=self.show_figures,
            backend=self.backend,
            save_location=temp_path,
        )

    @Magazine.reporting_figure(topic="Soil Moisture")
    def _soil_moisture_colour(self):
        """
        Implements colour soil moisture figure.
        """
        temp_path = self.temp_handler.store_figure(
            name="soil_moisture_colour_fig",
            topic=FigureTopic.SOIL_MOISTURE_VOL,
        )

        soil_moisture_coloured_figure(
            data_frame=self.data_frame,
            station_name=self.sensor_info.name,
            sm_column_name=str(ColumnInfo.Name.SOIL_MOISTURE_VOL_FINAL),
            save_location=temp_path,
        )

    @Magazine.reporting_figure(topic="Atmospheric Conditions")
    def _atmospheric_variables(self):
        """
        Implements 3 panel figure with atmopsheric variables
        """
        temp_path = self.temp_handler.store_figure(
            name="atmospheric_variables",
            topic=FigureTopic.ATMOSPHERIC,
        )
        temperature_min = (
            self.data_frame[str(ColumnInfo.Name.AIR_TEMPERATURE)].min() - 2
        )
        temperature_max = (
            self.data_frame[str(ColumnInfo.Name.AIR_TEMPERATURE)].max() + 2
        )

        atmospheric_conditions_figure(
            data_frame=self.data_frame,
            station_name=self.sensor_info.name,
            pressure_col=str(ColumnInfo.Name.AIR_PRESSURE),
            temperature_col=str(ColumnInfo.Name.AIR_TEMPERATURE),
            rel_humidity_col=str(ColumnInfo.Name.AIR_RELATIVE_HUMIDITY),
            temperature_range=(temperature_min, temperature_max),
            show=self.show_figures,
            backend=self.backend,
            save_location=temp_path,
        )

    @Magazine.reporting_figure(topic="Neutron Correction")
    def _neutron_correction_factors(self):
        """
        Implements the correction factor figure
        """
        temp_path = self.temp_handler.store_figure(
            name="correction_factors_figure",
            topic=FigureTopic.NEUTRONS,
        )
        correction_factors_figure(
            data_frame=self.data_frame,
            station_name=self.sensor_info.name,
            pressure_corr_col=str(ColumnInfo.Name.PRESSURE_CORRECTION),
            humidity_corr_col=str(ColumnInfo.Name.HUMIDITY_CORRECTION),
            intensity_corr_col=str(ColumnInfo.Name.INTENSITY_CORRECTION),
            biomass_corr_col=str(
                ColumnInfo.Name.ABOVEGROUND_BIOMASS_CORRECTION
            ),
            show=self.show_figures,
            backend=self.backend,
            save_location=temp_path,
        )

    def _create_intended_figures_list(self):
        """
        Create a list of the figures to be created.

        Returns
        -------
        List
            List of figures to be made
        """
        # Get list of figures to create
        if self.selected_figures:
            figures_to_process = [
                name
                for name, metadata in self._figure_registry.items()
                if name in self.selected_figures
                and metadata.topic not in self.ignore_sections
            ]
        elif self.create_all:
            figures_to_process = [
                name
                for name, metadata in self._figure_registry.items()
                if metadata.topic not in self.ignore_sections
            ]
        else:
            figures_to_process = []

        return figures_to_process

    def create_figures(self):
        """
        Creates the figures
        """
        figures_to_process = self._create_intended_figures_list()

        # Create figures
        for figure_name in figures_to_process:
            try:
                metadata = self._figure_registry[figure_name]
                self._validate_required_columns(metadata=metadata)
                method = self._resolve_method(metadata=metadata)
                method()

            except Exception as e:
                print(f"Could not create figure {figure_name}: {str(e)}")



import pandas as pd
import pytest
from neptoon.corrections.factory.build_corrections import (
    CorrectionBuilder,
    CorrectNeutrons,
    CorrectionFactory,
    CorrectionTheory,
    CorrectionType,
)
from neptoon.corrections import (
    IncomingIntensityCorrectionZreda2012,
    IncomingIntensityCorrectionHawdon2014,
    HumidityCorrectionRosolem2013,
)
from neptoon.config.configuration_input import SensorInfo
from neptoon.columns.column_information import ColumnInfo
from neptoon.corrections.factory.build_corrections import Correction


### Test Correction class
class WrongCorrection(Correction):
    """This should break"""

    pass


class MockCorrection(Correction):

    def __init__(
        self, correction_type: str, correction_factor_column_name: str
    ):
        super().__init__(correction_type, correction_factor_column_name)

    def apply(self, data_frame=pd.DataFrame):
        data_frame[self.correction_factor_column_name] = 1
        return data_frame


def test_incorrectly_applied_correction():
    """
    Test that abstract class prevents application without apply
    """

    with pytest.raises(TypeError):
        WrongCorrection("test")


def test_get_correction_factor_column_name():
    """
    Test abstract creation
    """

    correction = MockCorrection(
        correction_type="test_type",
        correction_factor_column_name="test_col_name",
    )

    assert correction.correction_type == "test_type"
    assert correction.get_correction_factor_column_name() == "test_col_name"


def test_apply_method():
    """
    Test apply method writes to the correct column
    """
    correction = MockCorrection(
        correction_type="test_type",
        correction_factor_column_name="test_col_name",
    )
    df = pd.DataFrame({"data": [1, 2, 3]})
    result_df = correction.apply(df)

    assert "test_col_name" in str(result_df.columns)


def test_abstract_class_instantiation():
    """
    Test cannot create Correction class directly
    """
    with pytest.raises(TypeError):
        Correction("test_type")


### Test CorrectionBuilder


class MockCorrection2(Correction):
    def __init__(
        self,
        correction_type: str,
        factor: float,
        correction_factor_column_name: str = "empty",
    ):
        super().__init__(correction_type, correction_factor_column_name)
        self.factor = factor

    def apply(self, data_frame: pd.DataFrame):
        data_frame[self.correction_factor_column_name] = self.factor
        return data_frame


def test_add_correction_to_builder():
    """Test adding a correction to the builder."""
    builder = CorrectionBuilder()
    correction = MockCorrection2("test", 1.5)
    builder.add_correction(correction)
    assert "test" in builder.corrections
    assert builder.corrections["test"] == correction


def test_add_invalid_correction():
    builder = CorrectionBuilder()
    builder.add_correction("not_a_correction")
    assert len(builder.corrections) == 0


def remove_correction_by_type():
    builder = CorrectionBuilder()
    correction = MockCorrection2("to_remove", 1.5)
    builder.add_correction(correction)
    assert "to_remove" in builder.corrections
    builder.remove_correction_by_type("to_remove")
    assert "to_remove" not in builder.corrections


def test_get_corrections_stored_in_builder():
    builder = CorrectionBuilder()
    correction1 = MockCorrection2("a_correction", 1.5)
    correction2 = MockCorrection2("another_correction", 2.0)
    builder.add_correction(correction1)
    builder.add_correction(correction2)
    corrections = list(builder.get_corrections())
    assert len(corrections) == 2
    assert correction1 in corrections
    assert correction2 in corrections


### Test CorrectNeutrons


@pytest.fixture
def sample_df():
    return pd.DataFrame(
        {
            str(ColumnInfo.Name.EPI_NEUTRON_COUNT_CPH): [100, 200, 300],
            str(ColumnInfo.Name.EPI_NEUTRON_COUNT_RAW): [100, 200, 300],
        }
    )


@pytest.fixture
def correction_builder():
    builder = CorrectionBuilder()
    builder.add_correction(MockCorrection2("test1", 1.5, "correction_1"))
    builder.add_correction(MockCorrection2("test2", 2.0, "correction_2"))
    return builder


def test_init(sample_df, correction_builder):
    """Test initialization of CorrectNeutrons."""
    corrector = CorrectNeutrons(sample_df, correction_builder)
    assert corrector.crns_data_frame.equals(sample_df)
    assert corrector.correction_builder == correction_builder
    assert corrector.correction_columns == []


def test_add_correction_to_corrector(sample_df, correction_builder):
    """Test adding a single correction."""
    corrector = CorrectNeutrons(sample_df, correction_builder)
    new_correction = MockCorrection2("test3", 1.2)
    corrector.add_correction(new_correction)
    assert "test3" in corrector.correction_builder.corrections


def test_add_complete_correction_builder(sample_df):
    """Test adding a whole new correction builder."""
    corrector = CorrectNeutrons(sample_df, CorrectionBuilder())
    new_builder = CorrectionBuilder()
    new_builder.add_correction(MockCorrection("new", 1.1))
    corrector.add_correction_builder(new_builder)
    assert corrector.correction_builder == new_builder


def test_create_correction_factors(sample_df, correction_builder):
    """Test creating correction factors."""
    corrector = CorrectNeutrons(sample_df, correction_builder)
    result_df = corrector.create_correction_factors(sample_df)
    assert "correction_1" in result_df.columns
    assert "correction_2" in result_df.columns
    assert (result_df["correction_1"] == 1.5).all()
    assert (result_df["correction_2"] == 2.0).all()


def test_create_corrected_neutron_column(sample_df, correction_builder):
    """Test creating the corrected neutron column."""
    corrector = CorrectNeutrons(sample_df, correction_builder)
    df_with_factors = corrector.create_correction_factors(sample_df)
    result_df = corrector.create_corrected_neutron_column(df_with_factors)
    assert (
        str(ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT) in result_df.columns
    )
    expected = (
        sample_df[str(ColumnInfo.Name.EPI_NEUTRON_COUNT_CPH)] * 1.5 * 2.0
    )
    assert (
        result_df[str(ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT)] == expected
    ).all()


def test_correct_neutrons(sample_df, correction_builder):
    """Test the full neutron correction process."""
    corrector = CorrectNeutrons(sample_df, correction_builder)
    result_df = corrector.correct_neutrons()
    assert (
        str(ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT) in result_df.columns
    )
    expected = (
        sample_df[str(ColumnInfo.Name.EPI_NEUTRON_COUNT_CPH)] * 1.5 * 2.0
    )
    assert (
        result_df[str(ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT)] == expected
    ).all()


def test_property_setters(sample_df, correction_builder):
    """Test property setters for crns_data_frame and
    correction_builder."""
    corrector = CorrectNeutrons(sample_df, correction_builder)

    new_df = pd.DataFrame({"new": [1, 2, 3]})
    corrector.crns_data_frame = new_df
    assert corrector.crns_data_frame.equals(new_df)

    new_builder = CorrectionBuilder()
    corrector.correction_builder = new_builder
    assert corrector.correction_builder == new_builder

    with pytest.raises(AttributeError):
        corrector.correction_builder = "not a builder"


@pytest.fixture
def site_information():
    site_information = SensorInfo(
        name="test",
        country="DEU",
        identifier="101",
        install_date=pd.to_datetime("14-01-2011", dayfirst=True),
        latitude=51.37,
        longitude=12.55,
        elevation=140,
        time_zone=1,
        reference_incoming_neutron_value=150,
        avg_dry_soil_bulk_density=1.4,
        avg_lattice_water=0.01,
        avg_soil_organic_carbon=0,
        site_cutoff_rigidity=2.94,
    )
    return site_information


def test_correction_factory_intensity():
    """
    Test correction factory selects the right correction in intensity
    """
    factory = CorrectionFactory()
    tmp_corr = factory.create_correction(
        CorrectionType.INCOMING_INTENSITY, CorrectionTheory.ZREDA_2012
    )

    assert isinstance(
        tmp_corr,
        IncomingIntensityCorrectionZreda2012,
    )
    assert tmp_corr.correction_factor_column_name is str(
        ColumnInfo.Name.INTENSITY_CORRECTION
    )
    assert tmp_corr.correction_type is CorrectionType.INCOMING_INTENSITY

    factory = CorrectionFactory()
    tmp_corr2 = factory.create_correction(
        CorrectionType.INCOMING_INTENSITY, CorrectionTheory.HAWDON_2014
    )
    assert isinstance(
        tmp_corr2,
        IncomingIntensityCorrectionHawdon2014,
    )


@pytest.fixture
def df_with_ref_monitor():
    df_with_ref_monitor = pd.DataFrame(
        {
            str(ColumnInfo.Name.REFERENCE_INCOMING_NEUTRON_VALUE): [
                500,
                500,
                500,
                500,
                500,
            ],
            str(ColumnInfo.Name.SITE_CUTOFF_RIGIDITY): [
                4.2,
                4.2,
                4.2,
                4.2,
                4.2,
            ],
            str(ColumnInfo.Name.INCOMING_NEUTRON_INTENSITY): [
                555,
                546,
                515,
                496,
                500,
            ],
            str(ColumnInfo.Name.REFERENCE_MONITOR_CUTOFF_RIGIDITY): [
                2.4,
                2.4,
                2.4,
                2.4,
                2.4,
            ],
        }
    )
    return df_with_ref_monitor


@pytest.fixture
def df_without_ref_monitor():
    df_without_ref_monitor = pd.DataFrame(
        {
            str(ColumnInfo.Name.REFERENCE_INCOMING_NEUTRON_VALUE): [
                500,
                500,
                500,
                500,
                500,
            ],
            str(ColumnInfo.Name.SITE_CUTOFF_RIGIDITY): [
                4.2,
                4.2,
                4.2,
                4.2,
                4.2,
            ],
            str(ColumnInfo.Name.INCOMING_NEUTRON_INTENSITY): [
                555,
                546,
                515,
                496,
                500,
            ],
        }
    )
    return df_without_ref_monitor


def test_correction_factory_intensity_hawdon(
    df_with_ref_monitor,
    df_without_ref_monitor,
):
    """Test hawdon method when ref given and not given"""

    factory = CorrectionFactory()
    tmp_corr = factory.create_correction(
        correction_type=CorrectionType.INCOMING_INTENSITY,
        correction_theory=CorrectionTheory.HAWDON_2014,
    )
    assert tmp_corr.correction_type is CorrectionType.INCOMING_INTENSITY
    assert tmp_corr.correction_factor_column_name is str(
        ColumnInfo.Name.INTENSITY_CORRECTION
    )


@pytest.fixture
def df_lat_and_elevation():
    df_lat_and_elevation = pd.DataFrame(
        {
            str(ColumnInfo.Name.REFERENCE_INCOMING_NEUTRON_VALUE): [
                500,
                500,
                500,
                500,
                500,
            ],
            str(ColumnInfo.Name.SITE_CUTOFF_RIGIDITY): [
                4.2,
                4.2,
                4.2,
                4.2,
                4.2,
            ],
            str(ColumnInfo.Name.INCOMING_NEUTRON_INTENSITY): [
                555,
                546,
                515,
                496,
                500,
            ],
            str(ColumnInfo.Name.LATITUDE): [
                21,
                21,
                21,
                21,
                21,
            ],
            str(ColumnInfo.Name.ELEVATION): [
                600,
                600,
                600,
                600,
                600,
            ],
            str(ColumnInfo.Name.NMDB_REFERENCE_STATION): [
                "JUNG",
                "JUNG",
                "JUNG",
                "JUNG",
                "JUNG",
            ],
        }
    )
    return df_lat_and_elevation


def test_correction_factory_intensity_mcjannet_desilets(
    df_lat_and_elevation,
):
    """Test McJannetDesilets method"""

    factory = CorrectionFactory()
    tmp_corr = factory.create_correction(
        correction_type=CorrectionType.INCOMING_INTENSITY,
        correction_theory=CorrectionTheory.MCJANNET_DESILETS_2023,
    )
    assert tmp_corr.correction_type is CorrectionType.INCOMING_INTENSITY
    assert tmp_corr.correction_factor_column_name is str(
        ColumnInfo.Name.INTENSITY_CORRECTION
    )
    df_lat_and_elevation_output = tmp_corr.apply(df_lat_and_elevation)
    assert (
        str(ColumnInfo.Name.INTENSITY_CORRECTION)
        in df_lat_and_elevation_output.columns
    )
    assert (
        str(ColumnInfo.Name.RC_CORRECTION_FACTOR)
        in df_lat_and_elevation_output.columns
    )


def test_correction_factory_intensity_mcjannet_desilets_error(
    df_with_ref_monitor,
):
    """Test McJannetDesilets method error (wrong inputs)"""

    factory = CorrectionFactory()
    tmp_corr = factory.create_correction(
        correction_type=CorrectionType.INCOMING_INTENSITY,
        correction_theory=CorrectionTheory.MCJANNET_DESILETS_2023,
    )
    assert tmp_corr.correction_type is CorrectionType.INCOMING_INTENSITY
    assert tmp_corr.correction_factor_column_name is str(
        ColumnInfo.Name.INTENSITY_CORRECTION
    )
    with pytest.raises(ValueError):
        df = tmp_corr.apply(df_with_ref_monitor)
        return df


def test_correction_factory_pressure():
    """
    Test correction factory selects the right correction humidity.
    """
    df = pd.DataFrame(
        {
            str(ColumnInfo.Name.AIR_PRESSURE): [1000, 990, 1010, 1001, 999],
            str(ColumnInfo.Name.MEAN_PRESSURE): [1000, 1000, 1000, 1000, 1000],
            str(ColumnInfo.Name.LATITUDE): [34, 34, 34, 34, 34],
            str(ColumnInfo.Name.ELEVATION): [100, 100, 100, 100, 100],
            str(ColumnInfo.Name.SITE_CUTOFF_RIGIDITY): [
                2.3,
                2.3,
                2.3,
                2.3,
                2.3,
            ],
        }
    )

    factory = CorrectionFactory()
    tmp_corr = factory.create_correction(
        correction_type=CorrectionType.PRESSURE
    )
    assert tmp_corr.correction_type is CorrectionType.PRESSURE
    assert tmp_corr.correction_factor_column_name is str(
        ColumnInfo.Name.PRESSURE_CORRECTION
    )
    assert str(ColumnInfo.Name.PRESSURE_CORRECTION) not in df.columns
    df = tmp_corr.apply(df)
    assert str(ColumnInfo.Name.PRESSURE_CORRECTION) in df.columns


def test_correction_factory_humidity():
    """
    Test correction factory selects the right correction for humidity

    Parameters
    ----------
    site_information : _type_
        _description_
    """
    df = pd.DataFrame(
        {
            str(ColumnInfo.Name.AIR_RELATIVE_HUMIDITY): [67, 70, 78, 76, 55],
            str(ColumnInfo.Name.AIR_TEMPERATURE): [
                21,
                24,
                22,
                21,
                22,
            ],
        }
    )

    factory = CorrectionFactory()
    tmp_corr = factory.create_correction(
        correction_type=CorrectionType.HUMIDITY
    )
    assert isinstance(
        tmp_corr,
        HumidityCorrectionRosolem2013,
    )

    df = tmp_corr.apply(df.copy())

    actual_vp_col = str(ColumnInfo.Name.ACTUAL_VAPOUR_PRESSURE)
    sat_vp_col = str(ColumnInfo.Name.SATURATION_VAPOUR_PRESSURE)
    abs_hum_col = str(ColumnInfo.Name.ABSOLUTE_HUMIDITY)
    corr_col = str(ColumnInfo.Name.HUMIDITY_CORRECTION)

    for col in (actual_vp_col, sat_vp_col, abs_hum_col):
        assert col in df.columns, f"Expected column '{col}' in output"

    assert corr_col in df.columns
    for col in (actual_vp_col, sat_vp_col, abs_hum_col, corr_col):
        assert df[col].notna().any(), f"Column '{col}' is all NaNs"
    assert pd.api.types.is_float_dtype(df[corr_col])


def test_humidity_correction_uniform_input_gives_constant_factor():
    """
    If you feed the same (RH, T) in every row, the resulting humidity‐correction
    factor should be identical (i.e. one unique value).
    """
    n = 5
    df = pd.DataFrame(
        {
            str(ColumnInfo.Name.AIR_RELATIVE_HUMIDITY): [60] * n,
            str(ColumnInfo.Name.AIR_TEMPERATURE): [22] * n,
        }
    )

    tmp_corr = HumidityCorrectionRosolem2013(
        reference_absolute_humidity_value=0.0
    )
    df_out = tmp_corr.apply(df.copy())

    corr_col = str(ColumnInfo.Name.HUMIDITY_CORRECTION)
    assert df_out[corr_col].nunique() == 1


def test_direct_humidity_correction_with_existing_abs_humidity(monkeypatch):
    """
    If a DataFrame already has 'absolute_humidity', then apply(…) should
    skip recomputing absolute humidity and just run humidity_correction_rosolem2013
    once per row. Monkey patching the internal humidity_correction_rosolem2013
    function will confirm it is called with exactly (abs_hum, reference_val).
    """
    df = pd.DataFrame(
        {
            str(ColumnInfo.Name.ABSOLUTE_HUMIDITY): [5.0, 10.0, 20.0],
            str(ColumnInfo.Name.AIR_RELATIVE_HUMIDITY): [0, 0, 0],
            str(ColumnInfo.Name.AIR_TEMPERATURE): [0, 0, 0],
        }
    )

    # Step 2: instantiate with a non‐zero reference, so we can check args:
    reference_val = 3.0
    hc = HumidityCorrectionRosolem2013(
        reference_absolute_humidity_value=reference_val
    )

    calls = []

    def fake_hum_corr(abs_hum, ref_val):
        calls.append((abs_hum, ref_val))
        return abs_hum * 0.1  # return something deterministic

    module_path = hc.__class__.__module__
    monkeypatch.setattr(
        f"{module_path}.humidity_correction_rosolem2013", fake_hum_corr
    )
    df_out = hc.apply(df.copy())

    assert list(df_out[str(ColumnInfo.Name.ABSOLUTE_HUMIDITY)]) == [
        5.0,
        10.0,
        20.0,
    ]

    corr_col = str(ColumnInfo.Name.HUMIDITY_CORRECTION)
    assert corr_col in df_out.columns
    assert calls == [
        (5.0, reference_val),
        (10.0, reference_val),
        (20.0, reference_val),
    ], print(calls)

    expected = [5.0 * 0.1, 10.0 * 0.1, 20.0 * 0.1]
    assert list(df_out[corr_col]) == expected



from pathlib import Path
import typer

from neptoon.workflow import ProcessWithConfig
from neptoon.config import ConfigurationManager

app = typer.Typer()


@app.callback(invoke_without_command=True)
def main(
    processing_config: str = typer.Option(
        None, "--processing", "-p", 
        help="Path to the processing configuration YAML file"
    ),
    sensor_config: str = typer.Option(
        None, "--sensor", "-s", 
        help="Path to the sensor configuration YAML file"
    ),
    verbose: bool = typer.Option(
        False, "--verbose", "-v", 
        help="Enable verbose output"
    ),
):
    """
    Process CRNS data using configuration files.

    
    Example

    -------
    
    neptoon -p /path/to/process.yaml -s /path/to/sensor.yaml
    """
    if processing_config and sensor_config:
        typer.secho("Processing the sensor data...", fg=typer.colors.GREEN, bold=True)
        process_data(processing_config, sensor_config, verbose)
    elif processing_config or sensor_config:
        typer.echo(typer.style("Error:",fg=typer.colors.RED, bold=True) +" Both processing and station configs are required")
        raise typer.Exit(code=1)
    else:
        typer.echo("Type " + typer.style("neptoon --help", fg=typer.colors.CYAN, bold=True) + " for help.")


def process_data(processing_config: str, sensor_config: str, verbose: bool):
    """
    Process the data using the supplied config file locations.
    """
    processing_config_path = Path(processing_config)
    sensor_config_path = Path(sensor_config)
    
    if not processing_config_path.exists():
        typer.echo(f"Error: Processing configuration file not found: {processing_config_path}")
        raise typer.Exit(code=1)
    
    if not sensor_config_path.exists():
        typer.echo(f"Error: Station configuration file not found: {sensor_config_path}")
        raise typer.Exit(code=1)
    
    config = ConfigurationManager()
    
    try:
        config.load_configuration(file_path=sensor_config_path)
        config.load_configuration(file_path=processing_config_path)
        config_processor = ProcessWithConfig(configuration_object=config)
        config_processor.run_full_process() # Add verbose into run full process later TODO
        if verbose:
            typer.echo("Processing completed successfully.")
    except Exception as e:
        typer.echo(f"Error during processing: {str(e)}")
        if verbose:
            import traceback
            traceback.print_exc()
        raise typer.Exit(code=1)

if __name__ == "__main__":
    app()


import pandas as pd
from neptoon.hub import CRNSDataHub
from neptoon.columns import ColumnInfo
from neptoon.config.configuration_input import SensorInfo
import pytest


@pytest.fixture
def sample_crns_data():
    return pd.DataFrame(
        {
            "date_time": pd.date_range(
                start="2023-01-01", periods=5, freq="h"
            ),
            "epithermal_neutrons_raw": [100, 110, 105, 115, 108],
            "epithermal_neutrons_cph": [100, 110, 105, 115, 108],
            "air_pressure": [1000, 1005, 1002, 998, 1001],
            "air_relative_humidity": [80, 75, 76, 65, 89],
            "air_temperature": [23, 24, 25, 23, 20],
        }
    ).set_index("date_time")


@pytest.fixture
def example_data_hub(sample_crns_data):
    return CRNSDataHub(crns_data_frame=sample_crns_data)


def test_crns_data_hub_initialization(sample_crns_data):
    """
    Assert that the data_hub is initialised correctly
    """
    data_hub = CRNSDataHub(crns_data_frame=sample_crns_data)
    assert isinstance(data_hub, CRNSDataHub)
    assert data_hub.crns_data_frame.equals(sample_crns_data)


@pytest.fixture
def example_sensor_information():
    site_information = SensorInfo(
        name="test",
        country="DEU",
        identifier="101",
        latitude=51.37,
        longitude=12.55,
        elevation=140,
        time_zone=1,
        install_date=pd.to_datetime("14/03/2015", dayfirst=True),
        reference_incoming_neutron_value=150,
        avg_dry_soil_bulk_density=1.4,
        avg_lattice_water=0.01,
        avg_soil_organic_carbon=0,
        # mean_pressure=900,
        site_cutoff_rigidity=2.94,
        site_biomass=1,
        n0=200,
    )
    return site_information


@pytest.fixture
def sample_crns_data_corrected():
    return pd.DataFrame(
        {
            "date_time": pd.date_range(
                start="2023-01-01", periods=5, freq="h"
            ),
            str(ColumnInfo.Name.EPI_NEUTRON_COUNT_RAW): [
                100,
                110,
                105,
                115,
                108,
            ],
            str(ColumnInfo.Name.EPI_NEUTRON_COUNT_FINAL): [
                100,
                110,
                105,
                115,
                108,
            ],
            str(ColumnInfo.Name.AIR_PRESSURE): [1000, 1005, 1002, 998, 1001],
            str(ColumnInfo.Name.ABSOLUTE_HUMIDITY): [10, 15, 15, 20, 10],
            str(ColumnInfo.Name.AIR_TEMPERATURE): [23, 24, 25, 23, 20],
            str(ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT_FINAL): [
                110,
                120,
                130,
                120,
                120,
            ],
            str(ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT_LOWER_COUNT): [
                100,
                110,
                120,
                110,
                110,
            ],
            str(ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT_UPPER_COUNT): [
                120,
                130,
                140,
                130,
                130,
            ],
        }
    ).set_index("date_time")


@pytest.fixture
def sample_hub_corrected(
    sample_crns_data_corrected, example_sensor_information
):
    return CRNSDataHub(
        crns_data_frame=sample_crns_data_corrected,
        sensor_info=example_sensor_information,
    )


def test_produce_soil_moisture_estimates_default(sample_hub_corrected):
    """
    Check if soil moisture column is added
    """
    sample_hub_corrected.produce_soil_moisture_estimates(
        n0=2000,
        dry_soil_bulk_density=1.4,
        lattice_water=0.01,
        soil_organic_carbon=0.01,
    )

    assert (
        str(ColumnInfo.Name.SOIL_MOISTURE_VOL)
        in sample_hub_corrected.crns_data_frame.columns
    )
    print(sample_hub_corrected.crns_data_frame)



from pathlib import Path
from typing import Union
from neptoon.logging import get_logger
from datetime import timedelta
import re

core_logger = get_logger()


def validate_and_convert_file_path(
    file_path: Union[str, Path, None],
    base: Union[str, Path] = "",
) -> Path:
    """
    Ensures that file paths are correctly parsed into pathlib.Path
    objects.

    Parameters
    ----------
    file_path : Union[str, Path]
        The path to the folder or file.

    Returns
    -------
    pathlib.Path
        The file_path as a pathlib.Path object.

    Raises
    ------
    ValueError
        Error if string, pathlib.Path, or None not given.
    """

    if file_path is None:
        return None
    if isinstance(file_path, str):
        new_file_path = Path(file_path)
        if new_file_path.is_absolute():
            return new_file_path
        else:
            if base == "":
                return Path.cwd() / Path(file_path)
            else:
                return base / Path(file_path)
    elif isinstance(file_path, Path):
        if file_path.is_absolute():
            return file_path
        else:
            if base == "":
                return Path.cwd() / Path(file_path)
            else:
                return base / file_path
    else:
        message = (
            "data_location must be of type str or pathlib.Path. \n"
            f"{type(file_path).__name__} provided, "
            "please change this."
        )
        core_logger.error(message)
        raise ValueError(message)


def parse_resolution_to_timedelta(
    resolution_str: str,
):
    """
    Parse a string representation of a time resolution and convert
    it to a timedelta object.

    This method takes a string describing a time resolution (e.g.,
    "30 minutes", "2 hours", "1 day") and converts it into a Python
    timedelta object. It supports minutes, hours, and days as units.

    Parameters
    ----------
    resolution_str : str
        A string representing the time resolution. The format should
        be "<number> <unit>", where <number> is a positive integer
        and <unit> is one of the following: - For minutes: "min",
        "minute", "minutes" - For hours: "hour", "hours", "hr",
        "hrs" - For days: "day", "days" The parsing is
        case-insensitive.

    Returns
    -------
    datetime.timedelta
        A timedelta object representing the parsed time resolution.

    Raises
    ------
    ValueError
        If the resolution string format is invalid or cannot be
        parsed.
    ValueError
        If an unsupported time unit is provided.
    """

    pattern = re.compile(r"(\d+)\s*([a-zA-Z]+)")
    match = pattern.match(resolution_str.strip())

    if not match:
        raise ValueError(f"Invalid resolution format: {resolution_str}")

    value, unit = match.groups()
    value = int(value)

    if unit.lower() in ["min", "mins", "minute", "minutes", "m"]:
        return timedelta(minutes=value)
    elif unit.lower() in ["hour", "hours", "hr", "hrs", "h"]:
        return timedelta(hours=value)
    elif unit.lower() in ["day", "days", "d"]:
        return timedelta(days=value)
    else:
        message = f"Unsupported time unit: {unit}"
        core_logger.error(message)
        raise ValueError(message)



from __future__ import annotations

import os
import platform
import re

__all__ = ['process_macos_plat_tag']


def process_macos_plat_tag(plat: str, /, *, compat: bool) -> str:
    """
    Process the macOS platform tag. This will normalize the macOS version to
    10.16 if compat=True. If the MACOSX_DEPLOYMENT_TARGET environment variable
    is set, then it will be used instead for the target version.  If archflags
    is set, then the archs will be respected, including a universal build.
    """
    # Default to a native build
    current_arch = platform.machine()
    arm = current_arch == 'arm64'

    # Look for cross-compiles
    archflags = os.environ.get('ARCHFLAGS', '')
    if archflags and (archs := re.findall(r'-arch (\S+)', archflags)):
        new_arch = 'universal2' if set(archs) == {'x86_64', 'arm64'} else archs[0]
        arm = archs == ['arm64']
        plat = f'{plat[: plat.rfind(current_arch)]}{new_arch}'

    # Process macOS version
    if sdk_match := re.search(r'macosx_(\d+_\d+)', plat):
        macos_version = sdk_match.group(1)
        target = os.environ.get('MACOSX_DEPLOYMENT_TARGET', None)

        try:
            new_version = normalize_macos_version(target or macos_version, arm=arm, compat=compat)
        except ValueError:
            new_version = normalize_macos_version(macos_version, arm=arm, compat=compat)

        return plat.replace(macos_version, new_version, 1)

    return plat


def normalize_macos_version(version: str, *, arm: bool, compat: bool) -> str:
    """
    Set minor version to 0 if major is 11+. Enforces 11+ if arm=True. 11+ is
    converted to 10.16 if compat=True. Version is always returned in
    "major_minor" format.
    """
    version = version.replace('.', '_')
    if '_' not in version:
        version = f'{version}_0'
    major, minor = (int(d) for d in version.split('_')[:2])
    major = max(major, 11) if arm else major
    minor = 0 if major >= 11 else minor  # noqa: PLR2004
    if compat and major >= 11:  # noqa: PLR2004
        major = 10
        minor = 16
    return f'{major}_{minor}'



from __future__ import annotations

import os

from hatchling.version.source.plugin.interface import VersionSourceInterface


class CodeSource(VersionSourceInterface):
    PLUGIN_NAME = 'code'

    def get_version_data(self) -> dict:
        import sys
        from importlib.util import module_from_spec, spec_from_file_location

        relative_path = self.config.get('path')
        if not relative_path:
            message = 'option `path` must be specified'
            raise ValueError(message)

        if not isinstance(relative_path, str):
            message = 'option `path` must be a string'
            raise TypeError(message)

        path = os.path.normpath(os.path.join(self.root, relative_path))
        if not os.path.isfile(path):
            message = f'file does not exist: {relative_path}'
            raise OSError(message)

        expression = self.config.get('expression') or '__version__'
        if not isinstance(expression, str):
            message = 'option `expression` must be a string'
            raise TypeError(message)

        search_paths = self.config.get('search-paths', [])
        if not isinstance(search_paths, list):
            message = 'option `search-paths` must be an array'
            raise TypeError(message)

        absolute_search_paths = []
        for i, search_path in enumerate(search_paths, 1):
            if not isinstance(search_path, str):
                message = f'entry #{i} of option `search-paths` must be a string'
                raise TypeError(message)

            absolute_search_paths.append(os.path.normpath(os.path.join(self.root, search_path)))

        spec = spec_from_file_location(os.path.splitext(path)[0], path)
        module = module_from_spec(spec)  # type: ignore[arg-type]

        old_search_paths = list(sys.path)
        try:
            sys.path[:] = [*absolute_search_paths, *old_search_paths]
            spec.loader.exec_module(module)  # type: ignore[union-attr]
        finally:
            sys.path[:] = old_search_paths

        # Execute the expression to determine the version
        version = eval(expression, vars(module))  # noqa: S307

        return {'version': version}

    def set_version(self, version: str, version_data: dict) -> None:  # noqa: ARG002, PLR6301
        message = 'Cannot rewrite loaded code'
        raise NotImplementedError(message)



import streamlit as st
from pathlib import Path
from magazine import Magazine

Magazine.active = True
# Icons: https://fonts.google.com/icons?icon.set=Material+Symbols&icon.style=Rounded
current_path = Path(__file__).parent

st.set_page_config(
    page_title="Neptoon GUI",
    page_icon="assets/neptoon-logo-ball.svg",
)

st.logo(
    image=current_path / "assets" / "neptoon-logo.svg",
    link="https://codebase.helmholtz.cloud/cosmos/neptoon_documentation",
    # icon_image=LOGO_URL_SMALL,
)

pages = [
    st.Page(
        "gui-configuration.py",
        title="Configuration",
        icon=":material/settings:",
    ),
    st.Page(
        "gui-site_information.py",
        title="Site information",
        icon=":material/info:",
    ),
    st.Page(
        "gui-read_data.py",
        title="Read data",
        icon=":material/full_stacked_bar_chart:",
    ),
    st.Page(
        "gui-neutron_corrections.py",
        title="Neutron corrections",
        icon=":material/blur_on:",
    ),
    st.Page(
        "gui-calibration.py",
        title="Calibration",
        icon=":material/adjust:",
    ),
    st.Page(
        "gui-water.py",
        title="Water",
        icon=":material/water_drop:",
    ),
    st.Page(
        "gui-run_all.py",
        title="Run all",
        icon=":material/web_traffic:",
    ),
    st.Page(
        "gui-export.py",
        title="Export",
        icon=":material/save:",
    ),
]

# Initialize sharable session variables
shared_session_variables = dict(
    config_sensor_selected="Custom configuration...",
    config_sensor_file=None,
    config_sensor_uploaded_file=None,
    config_sensor_uploaded_name="",
    config_processing_selected="COSMOS Standard v1.0",
    config_processing_file=None,
    config_processing_uploaded_file=None,
    config_processing_uploaded_name="",
    config_already_parsed=False,
    yaml=None,
    yaml_checked=False,
    data_preformatted_file=None,
    data_preformatted_upload_file=None,
    data_preformatted_upload_name="",
    data_preraw_file=None,
    data_preraw_upload_file=None,
    data_preraw_upload_name="",
    data_read_ready=False,
    data_parsed=False,
    data_nmdb_attached=False,
    data_quality_checked=False,
    data_corrections_made=False,
    calibration_file=None,
    calibration_upload_file=None,
    calibration_upload_name="",
    calibration_read_ready=False,
    calibration_finished=False,
    data_converted=False,
)
for var in shared_session_variables:
    if var not in st.session_state:
        st.session_state[var] = shared_session_variables[var]
    st.session_state[var] = st.session_state[var]  # keeps widget keys alive
# st.session_state[var] = "Custom configuration..."


st.sidebar.image(
    image=current_path / "assets" / "neptoon-affils.svg",
    use_container_width=False,
    width=70,
)


# lets turn on the reporting system for our data


pg = st.navigation(pages)
pg.run()



import pandas as pd
import pytest
from pathlib import Path
from unittest.mock import Mock
from neptoon.io.save import SaveAndArchiveOutputs
from neptoon.config.configuration_input import SensorInfo

UNFLAGGED = "UNFLAGGED"
BAD = "BAD"


@pytest.fixture
def sample_data():
    processed_df = pd.DataFrame(
        {
            "date_time": pd.date_range(
                start="2023-01-01", periods=5, freq="h"
            ),
            "epithermal_neutrons_raw": [100, 110, 105, 115, 108],
            "epithermal_neutrons_cph": [100, 110, 105, 115, 108],
            "air_pressure": [1000, 1005, 1002, 998, 1001],
            "air_relative_humidity": [80, 75, 76, 65, 89],
            "air_temperature": [23, 24, 25, 23, 20],
        }
    )
    flag_df = pd.DataFrame(
        {
            "date_time": pd.date_range(
                start="2023-01-01", periods=5, freq="h"
            ),
            "epithermal_neutrons_raw": [
                UNFLAGGED,
                UNFLAGGED,
                UNFLAGGED,
                BAD,
                BAD,
            ],
            "epithermal_neutrons_cph": [
                UNFLAGGED,
                BAD,
                BAD,
                UNFLAGGED,
                UNFLAGGED,
            ],
            "air_pressure": [
                UNFLAGGED,
                UNFLAGGED,
                UNFLAGGED,
                UNFLAGGED,
                UNFLAGGED,
            ],
            "air_relative_humidity": [
                UNFLAGGED,
                UNFLAGGED,
                BAD,
                UNFLAGGED,
                UNFLAGGED,
            ],
            "air_temperature": [
                UNFLAGGED,
                UNFLAGGED,
                UNFLAGGED,
                UNFLAGGED,
                UNFLAGGED,
            ],
        }
    )
    site_info = Mock(spec=SensorInfo)
    site_info.name = "TestSite"
    return processed_df, flag_df, site_info


@pytest.fixture
def save_and_archive(sample_data, tmp_path):
    processed_df, flag_df, site_info = sample_data
    return SaveAndArchiveOutputs(
        folder_name="test_folder",
        processed_data_frame=processed_df,
        flag_data_frame=flag_df,
        sensor_info=site_info,
        save_folder_location=tmp_path,
    )


def test_init(save_and_archive, tmp_path):
    """
    Tests intantiation of SaveAndArchiveOutputs
    """
    assert save_and_archive.folder_name == "test_folder"
    assert save_and_archive.save_folder_location == tmp_path
    assert save_and_archive.append_timestamp
    assert not save_and_archive.use_custom_column_names
    assert save_and_archive.custom_column_names_dict is None


def test_validate_save_folder(save_and_archive, tmp_path):
    """
    Tests that the folder name validation works as expected.
    """
    assert save_and_archive._validate_save_folder(tmp_path) == tmp_path
    assert save_and_archive._validate_save_folder(str(tmp_path)) == tmp_path
    assert save_and_archive._validate_save_folder(None) == Path.cwd()


def test_append_hash_to_folder_name(save_and_archive, tmp_path):
    """
    Test appending the hash to the folder.
    """
    folder_path = tmp_path / "test_folder"
    folder_path.mkdir()
    (folder_path / "data_audit_log").mkdir()
    (folder_path / "data_audit_log" / "unknown").mkdir()
    hash_file = folder_path / "data_audit_log" / "unknown" / "hash.txt"
    hash_file.write_text("123456abcdef")
    new_path = save_and_archive.append_hash_to_folder_name(folder_path)
    assert new_path == tmp_path / "test_folder_123456"
    assert new_path.exists()
    assert not folder_path.exists()


def test_mask_bad_data(save_and_archive):
    """
    Test data masking
    """
    masked_df = save_and_archive.mask_bad_data()
    assert pd.isna(masked_df["epithermal_neutrons_raw"][3])
    assert pd.isna(masked_df["epithermal_neutrons_cph"][1])
    assert pd.isna(masked_df["epithermal_neutrons_cph"][2])
    assert masked_df["epithermal_neutrons_raw"][0] == 100



from __future__ import annotations

from typing import Callable, TypeVar

import pluggy


class PluginManager:
    def __init__(self) -> None:
        self.manager = pluggy.PluginManager('hatch')
        self.third_party_plugins = ThirdPartyPlugins(self.manager)
        self.initialized = False

    def initialize(self) -> None:
        from hatchling.plugin import specs

        self.manager.add_hookspecs(specs)

    def __getattr__(self, name: str) -> ClassRegister:
        if not self.initialized:
            self.initialize()
            self.initialized = True

        hook_name = f'hatch_register_{name}'
        hook = getattr(self, hook_name, None)
        if hook:
            hook()

        register = ClassRegister(getattr(self.manager.hook, hook_name), 'PLUGIN_NAME', self.third_party_plugins)
        setattr(self, name, register)
        return register

    def hatch_register_version_source(self) -> None:
        from hatchling.version.source.plugin import hooks

        self.manager.register(hooks)

    def hatch_register_version_scheme(self) -> None:
        from hatchling.version.scheme.plugin import hooks

        self.manager.register(hooks)

    def hatch_register_builder(self) -> None:
        from hatchling.builders.plugin import hooks

        self.manager.register(hooks)

    def hatch_register_build_hook(self) -> None:
        from hatchling.builders.hooks.plugin import hooks

        self.manager.register(hooks)

    def hatch_register_metadata_hook(self) -> None:
        from hatchling.metadata.plugin import hooks

        self.manager.register(hooks)


class ClassRegister:
    def __init__(self, registration_method: Callable, identifier: str, third_party_plugins: ThirdPartyPlugins) -> None:
        self.registration_method = registration_method
        self.identifier = identifier
        self.third_party_plugins = third_party_plugins

    def collect(self, *, include_third_party: bool = True) -> dict:
        if include_third_party and not self.third_party_plugins.loaded:
            self.third_party_plugins.load()

        classes: dict[str, type] = {}

        for raw_registered_classes in self.registration_method():
            registered_classes = (
                raw_registered_classes if isinstance(raw_registered_classes, list) else [raw_registered_classes]
            )
            for registered_class in registered_classes:
                name = getattr(registered_class, self.identifier, None)
                if not name:  # no cov
                    message = f'Class `{registered_class.__name__}` does not have a {name} attribute.'
                    raise ValueError(message)

                if name in classes:  # no cov
                    message = (
                        f'Class `{registered_class.__name__}` defines its name as `{name}` but '
                        f'that name is already used by `{classes[name].__name__}`.'
                    )
                    raise ValueError(message)

                classes[name] = registered_class

        return classes

    def get(self, name: str) -> type | None:
        if not self.third_party_plugins.loaded:
            classes = self.collect(include_third_party=False)
            if name in classes:
                return classes[name]

        return self.collect().get(name)


class ThirdPartyPlugins:
    def __init__(self, manager: pluggy.PluginManager) -> None:
        self.manager = manager
        self.loaded = False

    def load(self) -> None:
        self.manager.load_setuptools_entrypoints('hatch')
        self.loaded = True


PluginManagerBound = TypeVar('PluginManagerBound', bound=PluginManager)



import pytest
import pandas as pd
from neptoon.quality_control.quality_assessment import (
    QAMethod,
    QATarget,
    ValidationError,
    QualityCheck,
    DateTimeIndexValidator,
    DataQualityAssessor,
)
from datetime import datetime
from neptoon.columns import ColumnInfo


@pytest.fixture
def df():
    df = pd.DataFrame(
        {
            "range": [
                400,
                500,
                400,
                300,
                500,
            ],
            "spike": [
                1010,
                2000,
                2300,
                2300,
                8000,
            ],
        }
    )
    return df


def test_quality_check_validation():

    with pytest.raises(ValidationError):
        QualityCheck(
            target=str(ColumnInfo.Name.INCOMING_NEUTRON_INTENSITY),
            method=QAMethod.RANGE_CHECK,
            parameters={"min": 500},
        )


def test_wrong_param_supplied():
    with pytest.raises(ValidationError):
        QualityCheck(
            target=str(ColumnInfo.Name.INCOMING_NEUTRON_INTENSITY),
            method=QAMethod.RANGE_CHECK,
            parameters={
                "min": 500,
                "max": 550,
                "crazy_param": 550,
            },
        )


def test_column_assignment_quality_check_1():
    check = QualityCheck(
        target=QATarget.AIR_PRESSURE,
        method=QAMethod.RANGE_CHECK,
        parameters={
            "min": 500,
            "max": 550,
        },
    )
    assert "column_name" in check.parameters.keys()


def test_column_assignment_quality_check_2():
    check1 = QualityCheck(
        target=QATarget.AIR_PRESSURE,
        method=QAMethod.RANGE_CHECK,
        parameters={
            "min": 500,
            "max": 550,
        },
    )
    assert check1.parameters["column_name"] == str(
        ColumnInfo.Name.AIR_PRESSURE
    )


def test_date_time_index_validator():
    """
    Tests the object which checks if a DataFrame has a datetime index.
    """
    data = {
        "Column1": [1, 2, 3, 4, 5],
        "Column2": ["A", "B", "C", "D", "E"],
        "Column3": [10.1, 20.2, 30.3, 40.4, 50.5],
    }
    df = pd.DataFrame(data)
    with pytest.raises(ValueError):
        DateTimeIndexValidator(df)


def test_date_time_index_validator_with_date():
    """
    Tests the object which checks if a DataFrame has a datetime index.
    """
    data = {
        "Column1": [1, 2, 3, 4, 5],
        "Column2": ["A", "B", "C", "D", "E"],
        "Column3": [10.1, 20.2, 30.3, 40.4, 50.5],
    }
    df = pd.DataFrame(data)
    date_range = pd.date_range(start="2023-01-01", periods=5, freq="D")
    df.index = date_range
    DateTimeIndexValidator(df)


@pytest.fixture
def test_df():

    start_date = datetime.now().replace(
        hour=0, minute=0, second=0, microsecond=0
    )
    date_range = pd.date_range(start=start_date, periods=5, freq="h")

    test_df = pd.DataFrame(
        {
            str(ColumnInfo.Name.AIR_PRESSURE): [555, 546, 515, 496, 500],
            str(ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT_FINAL): [
                56,
                60,
                70,
                60,
                45,
            ],
        },
        index=date_range,
    )
    return test_df


def test_quality_assessment(test_df):

    check1 = QualityCheck(
        target=QATarget.AIR_PRESSURE,
        method=QAMethod.RANGE_CHECK,
        parameters={
            "min": 500,
            "max": 550,
        },
    )

    qa = DataQualityAssessor(data_frame=test_df)
    qa.add_quality_check(check1)

    qa.apply_quality_assessment()

    result_df = qa.return_flags_data_frame()
    expected_flags = [
        "BAD",
        "UNFLAGGED",
        "UNFLAGGED",
        "BAD",
        "UNFLAGGED",
    ]
    actual_flags = result_df[str(ColumnInfo.Name.AIR_PRESSURE)].tolist()

    assert (
        actual_flags == expected_flags
    ), f"Expected {expected_flags}, but got {actual_flags}"


def test_quality_assessment_multi(test_df):

    check1 = QualityCheck(
        target=QATarget.AIR_PRESSURE,
        method=QAMethod.RANGE_CHECK,
        parameters={
            "min": 500,
            "max": 550,
        },
    )

    check2 = QualityCheck(
        target=QATarget.CORRECTED_EPI_NEUTRONS,
        method=QAMethod.RANGE_CHECK,
        parameters={
            "min": 50,
            "max": 65,
        },
    )

    qa = DataQualityAssessor(data_frame=test_df)
    qa.add_quality_check(check1)
    qa.add_quality_check(check2)

    qa.apply_quality_assessment()

    result_df = qa.return_flags_data_frame()
    print(result_df)
    expected_flags1 = [
        "BAD",
        "UNFLAGGED",
        "UNFLAGGED",
        "BAD",
        "UNFLAGGED",
    ]
    expected_flags2 = [
        "UNFLAGGED",
        "UNFLAGGED",
        "BAD",
        "UNFLAGGED",
        "BAD",
    ]
    actual_flag1 = result_df[str(ColumnInfo.Name.AIR_PRESSURE)].tolist()
    actual_flag2 = result_df[
        str(ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT_FINAL)
    ].tolist()

    assert (
        actual_flag1 == expected_flags1
    ), f"Expected {expected_flags1}, but got {actual_flag1}"

    assert (
        actual_flag2 == expected_flags2
    ), f"Expected {expected_flags2}, but got {actual_flag2}"


# def test_data_quality_assessor_scheme_change():
#     df = pd.DataFrame(
#         {"A": [1, 2, 3]}, index=pd.date_range("2023-01-01", periods=3)
#     )
#     dqa = DataQualityAssessor(data_frame=df)

#     # Test default SaQC scheme
#     assert dqa.saqc_scheme == "simple"

#     dqa.change_saqc_scheme("float")

#     # Test changing scheme
#     assert dqa.saqc_scheme == "float"

#     # Test wrong scheme given
#     with pytest.raises(TypeError):
#         dqa.change_saqc_scheme("wrong_scheme")



from neptoon.columns.column_information import ColumnInfo
import pytest


@pytest.fixture(autouse=True)
def reset_column_info():
    """
    This function will be called before every test function here.
    Since the test functions will change the original Class,
    it is necessary to set the class labels back to default after the tests.
    """
    # Reset the labels (probably not necessary)
    # ColumnInfo.reset_labels()
    # Run the test function
    yield
    # Reset labels again to the default class
    ColumnInfo.reset_labels()


def test_import_and_use_strings():
    """
    Test that the way we import in code pulls the string correctly
    """
    test_val1 = str(ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT)
    test_val2 = str(ColumnInfo.Name.SOIL_MOISTURE_VOL)
    assert isinstance(test_val1, str)
    assert isinstance(test_val2, str)


def test_change_column_name_in_pipeline():
    """
    Test relabeling works as expected
    """
    test_val1 = str(ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT)
    assert isinstance(test_val1, str)
    ColumnInfo.relabel(ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT, "new_name")
    assert str(ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT) == "new_name"


def test_reset_names():
    """
    Test resetting the labels to default
    """
    initial_value = str(ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT)
    ColumnInfo.relabel(ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT, "new_name")
    changed_value = str(ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT)
    ColumnInfo.reset_labels()
    reset_value = str(ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT)

    assert changed_value == "new_name"
    assert reset_value != "new_name"
    assert reset_value == initial_value



from enum import Enum, auto
import copy
from neptoon.logging import get_logger

core_logger = get_logger()


class ColumnInfo:
    """
    Used for storing information related to the cols in CRNS time series
    data. The names of columns are stored here providing a clean area to
    update column names and have this applied across the code base.

    TODO: Ideas for improvements:
        - Add information on whether a col is 'supplied' or 'calculated'
        - Use this information to organise order of 'calculated' columns
    """

    class Name(Enum):
        """
        All of the columns that neptoon could use during any of its
        methods. It provides a place to store relevant information about
        columns. A key aspect is it allows a user to change the expected
        column names during processing without needing to change the
        codebase. By storing as an Enum IDEs can offer autofill when
        typing ColumnInfo.Name....

        The user calls it by requesting the
        string representation of a column name:

        Example
        --------

        >>> crns_data_frame[str(ColumnInfo.Name.SOIL_MOISTURE)] = 1

        """

        DATE_TIME = auto()
        EPI_NEUTRON_COUNT_RAW = auto()
        EPI_NEUTRON_COUNT_CPH = auto()
        EPI_NEUTRON_COUNT_FINAL = auto()
        AIR_PRESSURE = auto()
        AIR_RELATIVE_HUMIDITY = auto()
        AIR_TEMPERATURE = auto()
        INCOMING_NEUTRON_INTENSITY = auto()
        SATURATION_VAPOUR_PRESSURE = auto()
        ACTUAL_VAPOUR_PRESSURE = auto()
        ABSOLUTE_HUMIDITY = auto()
        HUMIDITY_CORRECTION = auto()
        INTENSITY_CORRECTION = auto()
        PRESSURE_CORRECTION = auto()
        ABOVEGROUND_BIOMASS_CORRECTION = auto()
        CORRECTED_EPI_NEUTRON_COUNT = auto()
        CORRECTED_EPI_NEUTRON_COUNT_FINAL = auto()
        CORRECTED_EPI_NEUTRON_COUNT_UNCERTAINTY = auto()
        CORRECTED_EPI_NEUTRON_COUNT_UPPER_COUNT = auto()
        CORRECTED_EPI_NEUTRON_COUNT_LOWER_COUNT = auto()
        THERM_NEUTRON_COUNT_RAW = auto()
        THERM_NEUTRON_COUNT_CPH = auto()
        PRECIPITATION = auto()
        SOIL_MOISTURE_GRAV = auto()
        SOIL_MOISTURE_VOL = auto()
        SOIL_MOISTURE_VOL_FINAL = auto()
        SOIL_MOISTURE_UNCERTAINTY_VOL_UPPER = auto()
        SOIL_MOISTURE_UNCERTAINTY_VOL_LOWER = auto()
        SOIL_MOISTURE_MEASURMENT_DEPTH = auto()
        SOIL_MOISTURE_MEASUREMENT_RADIUS = auto()
        LATITUDE = auto()
        LONGITUDE = auto()
        ELEVATION = auto()
        REFERENCE_INCOMING_NEUTRON_VALUE = auto()
        SITE_CUTOFF_RIGIDITY = auto()
        REFERENCE_MONITOR_CUTOFF_RIGIDITY = auto()
        NMDB_REFERENCE_STATION = auto()
        RC_CORRECTION_FACTOR = auto()
        DRY_SOIL_BULK_DENSITY = auto()
        LATTICE_WATER = auto()
        SOIL_ORGANIC_CARBON = auto()
        MEAN_PRESSURE = auto()
        SITE_BIOMASS = auto()
        N0 = auto()
        BETA_COEFFICIENT = auto()
        CALIB_DEPTH_OF_SAMPLE = auto()
        CALIB_DISTANCE_TO_SENSOR = auto()
        CALIB_BULK_DENSITY = auto()
        CALIB_PROFILE_ID = auto()
        CALIB_SOIL_MOISTURE_GRAVIMETRIC = auto()
        CALIB_SOIL_ORGANIC_CARBON = auto()
        CALIB_LATTICE_WATER = auto()

        def __str__(self):
            return ColumnInfo._current_representation[self]

    _default_representation: dict["ColumnInfo.Name", str] = {
        Name.DATE_TIME: "date_time",
        Name.EPI_NEUTRON_COUNT_RAW: "epithermal_neutrons_raw",
        Name.EPI_NEUTRON_COUNT_CPH: "epithermal_neutrons_cph",
        Name.EPI_NEUTRON_COUNT_FINAL: "epithermal_neutrons_cph",
        Name.AIR_PRESSURE: "air_pressure",
        Name.AIR_RELATIVE_HUMIDITY: "air_relative_humidity",
        Name.AIR_TEMPERATURE: "air_temperature",
        Name.INCOMING_NEUTRON_INTENSITY: "incoming_neutron_intensity",
        Name.SATURATION_VAPOUR_PRESSURE: "saturation_vapour_pressure",
        Name.ACTUAL_VAPOUR_PRESSURE: "actual_vapour_pressure",
        Name.ABSOLUTE_HUMIDITY: "absolute_humidity",
        Name.HUMIDITY_CORRECTION: "humidity_correction",
        Name.INTENSITY_CORRECTION: "incoming_neutron_intensity_correction",
        Name.PRESSURE_CORRECTION: "atmospheric_pressure_correction",
        Name.ABOVEGROUND_BIOMASS_CORRECTION: "aboveground_biomass_correction",
        Name.CORRECTED_EPI_NEUTRON_COUNT: "corrected_epithermal_neutrons",
        Name.CORRECTED_EPI_NEUTRON_COUNT_FINAL: "corrected_epithermal_neutrons",
        Name.CORRECTED_EPI_NEUTRON_COUNT_UNCERTAINTY: "corrected_epithermal_neutrons_uncertainty",
        Name.CORRECTED_EPI_NEUTRON_COUNT_UPPER_COUNT: "corrected_epithermal_neutrons_upper_count",
        Name.CORRECTED_EPI_NEUTRON_COUNT_LOWER_COUNT: "corrected_epithermal_neutrons_lower_count",
        Name.THERM_NEUTRON_COUNT_RAW: "thermal_neutron_count_raw",
        Name.THERM_NEUTRON_COUNT_CPH: "thermal_neutron_count_cph",
        Name.PRECIPITATION: "precipitation",
        Name.SOIL_MOISTURE_GRAV: "soil_moisture_gravimetric",
        Name.SOIL_MOISTURE_VOL: "soil_moisture_volumetric",
        Name.SOIL_MOISTURE_VOL_FINAL: "soil_moisture_volumetric",  # updated to processed soil moisture
        Name.SOIL_MOISTURE_UNCERTAINTY_VOL_UPPER: "soil_moisture_vol_uncertainty_upper",
        Name.SOIL_MOISTURE_UNCERTAINTY_VOL_LOWER: "soil_moisture_vol_uncertainty_lower",
        Name.SOIL_MOISTURE_MEASURMENT_DEPTH: "crns_measurement_depth_cm",
        Name.SOIL_MOISTURE_MEASUREMENT_RADIUS: "crns_measurement_radius_m",
        Name.LATITUDE: "latitude",
        Name.LONGITUDE: "longitude",
        Name.ELEVATION: "elevation",
        Name.REFERENCE_INCOMING_NEUTRON_VALUE: "reference_incoming_neutron_value",
        Name.SITE_CUTOFF_RIGIDITY: "site_cutoff_rigidity",
        Name.REFERENCE_MONITOR_CUTOFF_RIGIDITY: "reference_monitor_cutoff_rigidity",
        Name.NMDB_REFERENCE_STATION: "nmdb_reference_station",
        Name.RC_CORRECTION_FACTOR: "rc_correction_factor",
        Name.DRY_SOIL_BULK_DENSITY: "dry_soil_bulk_density",
        Name.LATTICE_WATER: "lattice_water",
        Name.SOIL_ORGANIC_CARBON: "soil_organic_carbon",
        Name.MEAN_PRESSURE: "mean_pressure",
        Name.SITE_BIOMASS: "site_biomass",
        Name.N0: "n0",
        Name.BETA_COEFFICIENT: "beta_coefficient",
        Name.CALIB_DEPTH_OF_SAMPLE: "depth_of_sample",
        Name.CALIB_DISTANCE_TO_SENSOR: "distance_to_sensor",
        Name.CALIB_BULK_DENSITY: "bulk_density",
        Name.CALIB_PROFILE_ID: "profile_id",
        Name.CALIB_SOIL_MOISTURE_GRAVIMETRIC: "soil_moisture_gravimetric",
        Name.CALIB_SOIL_ORGANIC_CARBON: "soil_organic_carbon",
        Name.CALIB_LATTICE_WATER: "lattice_water",
    }

    _current_representation = copy.deepcopy(_default_representation)

    """
    SITE_INFO_TO_COLUMN_INFO is a mapping dictionary used when adding
    columns to the crns_data_frame using the SiteInformation class.

    See CRNSDataHub.prepare_static_values() for context.
    """
    SITE_INFO_TO_COLUMN_INFO = {
        "latitude": Name.LATITUDE,
        "longitude": Name.LONGITUDE,
        "elevation": Name.ELEVATION,
        "reference_incoming_neutron_value": Name.REFERENCE_INCOMING_NEUTRON_VALUE,
        "dry_soil_bulk_density": Name.DRY_SOIL_BULK_DENSITY,
        "lattice_water": Name.LATTICE_WATER,
        "soil_organic_carbon": Name.SOIL_ORGANIC_CARBON,
        "site_cutoff_rigidity": Name.SITE_CUTOFF_RIGIDITY,
        "mean_pressure": Name.MEAN_PRESSURE,
        "site_biomass": Name.SITE_BIOMASS,
        "n0": Name.N0,
        "beta_coefficient": Name.BETA_COEFFICIENT,
        "reference_monitor_cutoff_rigidity": Name.REFERENCE_MONITOR_CUTOFF_RIGIDITY,
    }

    @classmethod
    def relabel(cls, column_name: Name, new_label: str):
        """
        Class method which allows a user to change the expected string
        of a column type.

        Parameters
        ----------
        column_name : Name
            The Name of the column e.g., Name.EPI_NEUTRON_COUNT
        new_label : str
            A string that represents the new column name to expect
            throughout processing.
        """
        cls._current_representation[column_name] = new_label

    @classmethod
    def reset_labels(cls):
        """
        Class method to reset all the labels to default values supplied
        in neptoon.
        """
        cls._current_representation = copy.deepcopy(
            cls._default_representation
        )

    @classmethod
    def get_col_name(cls, column_name: str):
        """
        Method to return the string representation of the name of a
        Column

        Parameters
        ----------
        column_name : str
            The enum object from which the name is required

        Returns
        -------
        str
            The string representation of the name.
        """
        return str(getattr(cls.Name, column_name.upper()))



import pandas as pd
import numpy as np
import copy
from datetime import timedelta
from typing import Literal, List

# from scipy.optimize import minimize
from neptoon.columns import ColumnInfo
from neptoon.corrections import (
    Schroen2017,
    neutrons_to_total_grav_soil_moisture_desilets_etal_2010,
)
from neptoon.corrections.theory.neutrons_to_soil_moisture import (
    compute_n0_koehli_etal_2021,
)
from neptoon.data_prep.conversions import AbsoluteHumidityCreator


def _create_water_equiv_soc(soil_organic_carbon: float):
    return soil_organic_carbon * 0.555


class CalibrationConfiguration:
    """
    Configuration class for calibration steps
    """

    def __init__(
        self,
        hours_of_data_around_calib: int = 6,
        converge_accuracy: float = 0.01,
        neutron_conversion_method: Literal[
            "desilets_etal_2010", "koehli_etal_2021"
        ] = "desilets_etal_2010",
        calib_data_date_time_column_name: str = str(ColumnInfo.Name.DATE_TIME),
        calib_data_date_time_format: str = "%Y-%m-%d %H:%M",
        sample_depth_column: str = str(ColumnInfo.Name.CALIB_DEPTH_OF_SAMPLE),
        distance_column: str = str(ColumnInfo.Name.CALIB_DISTANCE_TO_SENSOR),
        bulk_density_of_sample_column: str = str(
            ColumnInfo.Name.CALIB_BULK_DENSITY
        ),
        profile_id_column: str = str(ColumnInfo.Name.CALIB_PROFILE_ID),
        soil_moisture_gravimetric_column: str = str(
            ColumnInfo.Name.CALIB_SOIL_MOISTURE_GRAVIMETRIC
        ),
        soil_organic_carbon_column: str = str(
            ColumnInfo.Name.CALIB_SOIL_ORGANIC_CARBON
        ),
        lattice_water_column: str = str(ColumnInfo.Name.CALIB_LATTICE_WATER),
        abs_air_humidity_column: str = str(ColumnInfo.Name.ABSOLUTE_HUMIDITY),
        neutron_column_name: str = str(
            ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT_FINAL
        ),
        air_pressure_column_name: str = str(ColumnInfo.Name.AIR_PRESSURE),
        value_avg_lattice_water: float = 0,
        value_avg_bulk_density: float = 0,
        value_avg_soil_organic_carbon: float = 0,
        koehli_method_form: Literal[
            "Jan23_uranos",
            "Jan23_mcnpfull",
            "Mar12_atmprof",
            "Mar21_mcnp_drf",
            "Mar21_mcnp_ewin",
            "Mar21_uranos_drf",
            "Mar21_uranos_ewin",
            "Mar22_mcnp_drf_Jan",
            "Mar22_mcnp_ewin_gd",
            "Mar22_uranos_drf_gd",
            "Mar22_uranos_ewin_chi2",
            "Mar22_uranos_drf_h200m",
            "Aug08_mcnp_drf",
            "Aug08_mcnp_ewin",
            "Aug12_uranos_drf",
            "Aug12_uranos_ewin",
            "Aug13_uranos_atmprof",
            "Aug13_uranos_atmprof2",
        ] = "Mar21_uranos_drf",
    ):
        """
        Attributes.

        Parameters
        ----------
        hours_of_data_around_calib : int, optional
            Number of hours of neutron count data to include around the
            datetime stamp for calibration. This window is used to
            gather measurements from sensors during the calibration
            period. Default is 6.
        converge_accuracy : float, optional
            The convergence threshold for when finding n0. Default is
            0.01.
        neutron_conversion_method : {"desilets_etal_2010",
        "koehli_etal_2021"}, optional
            The conversion method used to translate raw neutron counts
            into soil moisture estimates. Options are
            "desilets_etal_2010" or "koehli_etal_2021". Default is
            "desilets_etal_2010".
        calib_data_date_time_column_name : str, optional
            The name of the column containing date‐time information for
            each calibration day. By default, this is set to
            str(ColumnInfo.Name.DATE_TIME).
        sample_depth_column : str, optional
            The name of the column with sample depth values (cm), by
            default str(ColumnInfo.Name.CALIB_DEPTH_OF_SAMPLE)
        distance_column : str, optional
            The name of the column stating the distance of the sample
            from the sensor (meters), by default
            str(ColumnInfo.Name.CALIB_DISTANCE_TO_SENSOR)
        bulk_density_of_sample_column : str, optional
            The name of the column with bulk density values of the
            samples (g/cm^3), by default str(
            ColumnInfo.Name.CALIB_BULK_DENSITY )
        profile_id_column : str, optional
            Name of the column with profile IDs, by default
            str(ColumnInfo.Name.CALIB_PROFILE_ID)
        soil_moisture_gravimetric_column : str, optional
            Name of the column with gravimetric soil moisture values
            (g/g), by default str(
            ColumnInfo.Name.CALIB_SOIL_MOISTURE_GRAVIMETRIC )
        soil_organic_carbon_column : str, optional
            Name of the column with soil organic carbon values (g/g), by
            default str( ColumnInfo.Name.CALIB_SOIL_ORGANIC_CARBON )
        lattice_water_column : str, optional
            Name of the column with lattice water values (g/g), by
            default str(ColumnInfo.Name.CALIB_LATTICE_WATER)
        abs_air_humidity_column : str, optional
            Name of the column with absolute air humidity values
            (g/cm3), by default str(ColumnInfo.Name.ABSOLUTE_HUMIDITY)
        neutron_column_name : str, optional
            Name of the column with corrected neutrons in it, by default
            str(ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT_FINAL)
        air_pressure_column_name : str, optional
            Name of the column with air pressure vlaues in it, by
            default str(ColumnInfo.Name.AIR_PRESSURE)
        value_avg_lattice_water: float, optional
            The actual site average lattice water value
        value_avg_bulk_density: float, optional
            The actual site average dry soil bulk density
        value_avg_soil_organic_carbon: float, optional
            The actual site average soil organic carbon
        """
        self.hours_of_data_around_calib = hours_of_data_around_calib
        self.converge_accuracy = converge_accuracy
        self.calib_data_date_time_column_name = (
            calib_data_date_time_column_name
        )
        self.neutron_conversion_method = neutron_conversion_method
        self.calib_data_date_time_format = calib_data_date_time_format
        self.sample_depth_column = sample_depth_column
        self.distance_column = distance_column
        self.bulk_density_of_sample_column = bulk_density_of_sample_column
        self.profile_id_column = profile_id_column
        self.soil_moisture_gravimetric_column = (
            soil_moisture_gravimetric_column
        )
        self.soil_organic_carbon_column = soil_organic_carbon_column
        self.lattice_water_column = lattice_water_column
        self.abs_air_humidity_column = abs_air_humidity_column
        self.neutron_column_name = neutron_column_name
        self.air_pressure_column_name = air_pressure_column_name
        self.value_avg_lattice_water = value_avg_lattice_water
        self.value_avg_bulk_density = value_avg_bulk_density
        self.value_avg_soil_organic_carbon = value_avg_soil_organic_carbon
        self.value_avg_soil_organic_carbon_water_equiv = (
            _create_water_equiv_soc(value_avg_soil_organic_carbon)
        )
        self.koehli_method_form = koehli_method_form


class CalibrationStation:
    """
    Abstract which does the complete claibration steps. Can be used on
    its own, but is mainly designed to facilitate CRNSDataHub
    calibration. Simply include the calibration data, the time series
    data and the config object and run find_n0_value(), to return the
    optimum N0.
    """

    def __init__(
        self,
        calibration_data: pd.DataFrame,
        time_series_data: pd.DataFrame,
        config: CalibrationConfiguration,
    ):
        self.calibration_data = calibration_data
        self.time_series_data = time_series_data
        self.config = config
        # place holders
        self.calib_prepper = None
        self.times_series_prepper = None
        self.calibrator = None

    def _collect_stats_for_magazine(self):
        self.number_calib_days = len(
            self.calibrator.return_output_dict_as_dataframe()
        )

    def find_n0_value(self):
        """
        Runs the full process to obtain an N0 estimate.

        Returns
        -------
        float
            N0 estimate after calibration.
        """
        self.calib_prepper = PrepareCalibrationData(
            calibration_data_frame=self.calibration_data,
            config=self.config,
        )
        self.calib_prepper.prepare_calibration_data()
        times_series_prepper = PrepareNeutronCorrectedData(
            corrected_neutron_data_frame=self.time_series_data,
            calibration_data_prepper=self.calib_prepper,
            config=self.config,
        )
        times_series_prepper.extract_calibration_day_values()
        self.calibrator = CalibrationWeightsCalculator(
            time_series_data_object=times_series_prepper,
            calib_data_object=self.calib_prepper,
            config=self.config,
        )
        self.calibrator.apply_weighting_to_multiple_days()
        optimal_n0 = self.calibrator.find_optimal_N0()
        return optimal_n0

    def return_calibration_results_data_frame(self):
        """
        Returns the daily results as a data frame. When multiple days
        calibration is undertaken on each day. The outputs of this are
        saved and this method returns them for viewing.

        Returns
        -------
        pd.DataFrame
            data frame with the results in it.
        """
        return self.calibrator.return_output_dict_as_dataframe()


class SampleProfile:

    latest_pid = 0

    def __init__(
        self,
        soil_moisture_gravimetric,
        depth,
        bulk_density,
        site_avg_bulk_density,
        site_avg_organic_carbon,
        site_avg_lattice_water,
        calibration_day,
        distance=1,
        lattice_water=None,
        soil_organic_carbon=None,
        pid=None,
    ):
        """
        Initialise SampleProfile instance.

        Parameters
        ----------
        soil_moisture_gravimetric : array
            array of soil moisture gravimetric values in g/g
        depth : array
            The depth of each soil moisture sample
        bulk_density : array
            bulk density of the samples in g/cm^3
        distance : int, optional
            distance of the profile from the sensor, by default 1
        lattice_water : array-like, optional
            Lattice water from the samples , by default 0
        soil_organic_carbon : int, optional
            _description_, by default 0
        pid : _type_, optional
            _description_, by default None
        """

        # Vector data
        if pid is None:
            SampleProfile.latest_pid += 1
            self.pid = SampleProfile.latest_pid
        else:
            self.pid = pid

        self.soil_moisture_gravimetric = np.array(soil_moisture_gravimetric)
        self.depth = np.array(depth)
        self.bulk_density = np.array(bulk_density)
        self.site_avg_bulk_density = site_avg_bulk_density
        self.calibration_day = calibration_day
        self.soil_organic_carbon = (
            np.array(soil_organic_carbon)
            if soil_organic_carbon is not None
            else np.zeros_like(soil_moisture_gravimetric)
        )
        self.site_avg_organic_carbon = site_avg_organic_carbon
        self.site_avg_organic_carbon_water_equiv = _create_water_equiv_soc(
            site_avg_organic_carbon
        )
        self.lattice_water = self.lattice_water = (
            np.array(lattice_water)
            if lattice_water is not None
            else np.zeros_like(soil_moisture_gravimetric)
        )
        self.site_avg_lattice_water = site_avg_lattice_water
        self.vertical_weights = np.ones_like(soil_moisture_gravimetric)
        self._calculate_sm_total_vol()
        self._calculate_sm_total_grv()

        self._distance = distance
        self.rescaled_distance = distance  # initialise as distance first
        self.D86 = np.nan
        self.sm_total_weighted_avg_grv = np.nan
        self.sm_total_weighted_avg_vol = np.nan
        self.horizontal_weight = 1  # intialise as 1

    @property
    def distance(self):
        return self._distance

    def _calculate_sm_total_vol(self):
        """
        Calculate total volumetric soil moisture.
        """
        # replace nan entries in bulk_density by the site average
        bd = np.where(
            np.isnan(self.bulk_density),
            self.site_avg_bulk_density,
            self.bulk_density,
        )
        self.sm_total_vol = self.soil_moisture_gravimetric * bd

    def _calculate_sm_total_grv(self):
        """
        Calculate total gravimetric soil moisture.
        """

        self.sm_total_grv = self.soil_moisture_gravimetric


class PrepareCalibrationData:
    """
    Prepares the calibration dataframe for processing.

    - ensures datetime index
    - splits multiple days into individual data frames
    - for each calibration day, converts data into a list of
      SampleProfiles
    - calculates site averages of key information (e.g., bulk density)
    - gap fills key info with averages when missing
    """

    def __init__(
        self,
        calibration_data_frame: pd.DataFrame,
        config: CalibrationConfiguration,
    ):
        """
        Instantiate attributes

        Parameters
        ----------
        calibration_data_frame : pd.DataFrame
            The dataframe with the calibration sample data in it. If
            multiple calibration days are available these should be
            stacked in the same dataframe.
        """

        self.calibration_data_frame = calibration_data_frame
        self.config = config
        self._ensure_date_time_index()

        self.unique_calibration_days = np.unique(
            self.calibration_data_frame[
                self.config.calib_data_date_time_column_name
            ]
        )
        self.list_of_data_frames = []
        self.list_of_profiles = []

    def _ensure_date_time_index(self):
        """
        Converts the date time column so the values are datetime type.
        """

        self.calibration_data_frame[
            self.config.calib_data_date_time_column_name
        ] = pd.to_datetime(
            self.calibration_data_frame[
                self.config.calib_data_date_time_column_name
            ],
            utc=True,
            dayfirst=True,
            format=self.config.calib_data_date_time_format,
        )

    def _create_list_of_df(self):
        """
        Splits up the self.calibration_data_frame into individual data
        frames, where each data frame is a different calibration day.
        """

        self.list_of_data_frames = [
            self.calibration_data_frame[
                self.calibration_data_frame[
                    self.config.calib_data_date_time_column_name
                ]
                == calibration_day
            ]
            for calibration_day in self.unique_calibration_days
        ]

    def _create_calibration_day_profiles(
        self,
        single_day_data_frame,
        site_avg_bulk_density,
        site_avg_lattice_water,
        site_avg_organic_carbon,
    ):
        """
        Returns a list of SampleProfile objects which have been created
        from a single calibration day data frame.

        Parameters
        ----------
        single_day_data_frame : pd.DataFrame
            _description_

        Returns
        -------
        List of SampleProfiles
            A list of created SampleProfiles
        """
        calibration_day_profiles = []
        profile_ids = np.unique(
            single_day_data_frame[self.config.profile_id_column]
        )
        for pid in profile_ids:
            temp_df = single_day_data_frame[
                single_day_data_frame[self.config.profile_id_column] == pid
            ]
            soil_profile = self._create_individual_profile(
                pid=pid,
                profile_data_frame=temp_df,
                site_avg_bulk_density=site_avg_bulk_density,
                site_avg_lattice_water=site_avg_lattice_water,
                site_avg_organic_carbon=site_avg_organic_carbon,
            )

            calibration_day_profiles.append(soil_profile)
        return calibration_day_profiles

    def _create_individual_profile(
        self,
        pid,
        profile_data_frame,
        site_avg_bulk_density,
        site_avg_lattice_water,
        site_avg_organic_carbon,
    ):
        """
        Creates a SampleProfile object from a individual profile
        dataframe

        Parameters
        ----------
        pid : numeric
            The profile ID to represent the profile.
        profile_data_frame : pd.DataFrame
            A data frame which holds the values for one single profile.

        Returns
        -------
        SampleProfile
            A SampleProfile object is returned.
        """
        distances = profile_data_frame[self.config.distance_column].median()
        depths = profile_data_frame[self.config.sample_depth_column]
        bulk_density = profile_data_frame[
            self.config.bulk_density_of_sample_column
        ]

        soil_moisture_gravimetric = profile_data_frame[
            self.config.soil_moisture_gravimetric_column
        ]

        soil_organic_carbon = profile_data_frame[
            self.config.soil_organic_carbon_column
        ]

        lattice_water = profile_data_frame[self.config.lattice_water_column]
        # only need one calibration datetime
        calibration_datetime = profile_data_frame[
            self.config.calib_data_date_time_column_name
        ].iloc[0]
        soil_profile = SampleProfile(
            soil_moisture_gravimetric=soil_moisture_gravimetric,
            depth=depths,
            bulk_density=bulk_density,
            site_avg_bulk_density=site_avg_bulk_density,
            distance=distances,
            lattice_water=lattice_water,
            soil_organic_carbon=soil_organic_carbon,
            pid=pid,
            calibration_day=calibration_datetime,
            site_avg_lattice_water=site_avg_lattice_water,
            site_avg_organic_carbon=site_avg_organic_carbon,
        )
        return soil_profile

    def _create_site_avg_bulk_density(self):
        """
        Produces a average bulk density from provided sample data, if
        not available uses user provided average, if thats not available
        raises error

        Returns
        -------
        float
            site average bulk density

        Raises
        ------
        ValueError
            No bulk density provided
        """
        if (
            self.config.bulk_density_of_sample_column
            not in self.calibration_data_frame.columns
        ):
            message = (
                "There appears to be no bulk density data in the calibration sample dataset. "
                "Attempting to use the provided site average..."
            )
            print(message)
            if self.config.value_avg_bulk_density is None:
                message = (
                    "There is no provided site average bulk density value. Please provide this "
                    "in sensor configuration files (if using configs), SensorInformation (if using notebooks)"
                    "or CalibrationConfiguration (if using calibration module directly.)"
                )
                raise ValueError(message)
            else:
                message = "Using the provided site average bulk density value."
                print(message)

        else:
            message = "Calculating site average bulk density from provided sample data."
            print(message)
            return self.calibration_data_frame[
                self.config.bulk_density_of_sample_column
            ].mean()

    def _create_site_avg_lattice_water(self):
        """
        Produces a average lattice water from provided sample data, if
        not available uses user provided average, if thats not available
        raises error

        Returns
        -------
        float
            site average lattice water

        Raises
        ------
        ValueError
            No lattice water provided
        """
        if (
            self.config.lattice_water_column
            not in self.calibration_data_frame.columns
        ):
            message = (
                "There appears to be no lattice water data in the calibration sample dataset. "
                "Attempting to use the provided site average..."
            )
            print(message)
            if self.config.value_avg_lattice_water is None:
                message = (
                    "There is no provided site average lattice water value. Please provide this "
                    "in sensor configuration files (if using configs), SensorInformation (if using notebooks)"
                    "or CalibrationConfiguration (if using calibration module directly.)\n"
                    "NOTE: If you don't know, use 0"
                )
                raise ValueError(message)
            else:
                message = (
                    "Using the provided site average lattice water value."
                )
                print(message)

        else:
            message = "Calculating site average lattice water from provided sample data."
            print(message)
            return self.calibration_data_frame[
                self.config.lattice_water_column
            ].mean()

    def _create_site_avg_soil_organic_carbon(self):
        """
        Produces a average soil_organic_carbon from provided sample
        data, if not available uses user provided average, if thats not
        available raises error

        Returns
        -------
        float
            site average soil_organic_carbon

        Raises
        ------
        ValueError
            No soil_organic_carbon provided
        """
        if (
            self.config.soil_organic_carbon_column
            not in self.calibration_data_frame.columns
        ):
            message = (
                "There appears to be no soil_organic_carbon data in the calibration sample dataset. "
                "Attempting to use the provided site average..."
            )
            print(message)
            if self.config.value_avg_soil_organic_carbon is None:
                message = (
                    "There is no provided site average soil_organic_carbon value. Please provide this "
                    "in sensor configuration files (if using configs), SensorInformation (if using notebooks)"
                    "or CalibrationConfiguration (if using calibration module directly.)\n"
                    "NOTE: If you don't know, use 0"
                )
                raise ValueError(message)
            else:
                message = "Using the provided site average soil_organic_carbon value."
                print(message)

        else:
            message = "Calculating site average soil_organic_carbon from provided sample data."
            print(message)

            return self.calibration_data_frame[
                self.config.soil_organic_carbon_column
            ].mean()

    def prepare_calibration_data(self):
        """
        Prepares the calibration data into a list of profiles.
        """

        self.config.value_avg_bulk_density = (
            self._create_site_avg_bulk_density()
        )

        self.config.value_avg_lattice_water = (
            self._create_site_avg_lattice_water()
        )

        self.config.value_avg_soil_organic_carbon = (
            self.calibration_data_frame[
                self.config.soil_organic_carbon_column
            ].mean()
        )

        if np.isnan(self.config.value_avg_soil_organic_carbon):
            self.config.value_avg_soil_organic_carbon = 0

        self._create_list_of_df()

        for data_frame in self.list_of_data_frames:
            calibration_day_profiles = self._create_calibration_day_profiles(
                single_day_data_frame=data_frame,
                site_avg_bulk_density=self.config.value_avg_bulk_density,
                site_avg_organic_carbon=self.config.value_avg_soil_organic_carbon,
                site_avg_lattice_water=self.config.value_avg_lattice_water,
            )
            self.list_of_profiles.extend(calibration_day_profiles)


class PrepareNeutronCorrectedData:

    def __init__(
        self,
        corrected_neutron_data_frame: pd.DataFrame,
        calibration_data_prepper: PrepareCalibrationData,
        config: CalibrationConfiguration,
    ):
        self.corrected_neutron_data_frame = corrected_neutron_data_frame
        self.calibration_data_prepper = calibration_data_prepper
        self.config = config
        self.data_dict = {}

        self._ensure_date_time_index()
        self._ensure_abs_humidity_exists()

    def _ensure_date_time_index(self):
        """
        Converts the date time column so the values are datetime type.
        """

        self.corrected_neutron_data_frame.index = pd.to_datetime(
            self.corrected_neutron_data_frame.index,
            utc=True,
        )

    def _ensure_abs_humidity_exists(self):
        """
        Checks to see if absolute humidity exists in the data frame. If
        it doesn't it will create it.
        """
        if (
            str(ColumnInfo.Name.ABSOLUTE_HUMIDITY)
            not in self.corrected_neutron_data_frame.columns
        ):
            abs_humidity_creator = AbsoluteHumidityCreator(
                self.corrected_neutron_data_frame
            )
            self.corrected_neutron_data_frame = (
                abs_humidity_creator.check_and_return_abs_hum_column()
            )

    def extract_calibration_day_values(self):
        """
        Extracts the rows of data for each calibration day.
        """
        calibration_indicies_dict = self._extract_calibration_day_indices(
            hours_of_data=self.config.hours_of_data_around_calib
        )
        dict_of_data = {}
        for value in calibration_indicies_dict.values():
            tmp_df = self.corrected_neutron_data_frame.loc[value]
            calib_day = None
            # Find calibration day index to use as dict key
            for day in self.calibration_data_prepper.unique_calibration_days:
                calib_day = self._find_nearest_calib_day_in_indicies(
                    day=day, data_frame=tmp_df
                )
                if calib_day is not None:
                    break
            dict_of_data[calib_day] = tmp_df

        self.data_dict = dict_of_data

    def _find_nearest_calib_day_in_indicies(self, day, data_frame):

        day = pd.to_datetime(day)
        mask = (data_frame.index >= day - timedelta(hours=1)) & (
            data_frame.index <= day + timedelta(hours=1)
        )
        if mask.any():
            calib_day = day
            return calib_day

    def _extract_calibration_day_indices(
        self,
        hours_of_data=6,
    ):
        """
        Extracts the required indices

        Parameters
        ----------
        hours_of_data : int, optional
            The hours of data around the calibration time stampe to
            collect, by default 6

        Returns
        -------
        dict
            A dictionary for each calibration date with the indices to
            extract from corrected neutron data.
        """
        extractor = IndicesExtractor(
            corrected_neutron_data_frame=self.corrected_neutron_data_frame,
            calibration_data_prepper=self.calibration_data_prepper,
            hours_of_data_to_extract=hours_of_data,
        )
        calibration_indices = extractor.extract_calibration_day_indices()

        return calibration_indices


class IndicesExtractor:
    """
    Extracts indices from the corrected neutron data based on the
    supplied calibration days
    """

    def __init__(
        self,
        corrected_neutron_data_frame,
        calibration_data_prepper,
        hours_of_data_to_extract=6,
    ):
        """
        Attributes.

        Parameters
        ----------
        corrected_neutron_data_frame : pd.DataFrame
            The corrected neutron data frame
        calibration_data_prepper : PrepareCalibrationData
            The processed object
        hours_of_data_to_extract : int, optional
            The number of hours of data around the calibration date time
            stamp to collect., by default 6
        """
        self.corrected_neutron_data_frame = corrected_neutron_data_frame
        self.calibration_data_prepper = calibration_data_prepper
        self.hours_of_data_to_extract = hours_of_data_to_extract

    def _convert_to_datetime(
        self,
        dates,
    ):
        """
        Convert a list of dates to pandas Timestamp objects.
        """
        return pd.to_datetime(dates)

    def _create_time_window(
        self,
        date: pd.Timestamp,
    ):
        """
        Create a time window around a given date.
        """
        half_window = self.hours_of_data_to_extract / 2
        window = pd.Timedelta(hours=half_window)
        return date - window, date + window

    def _extract_indices_within_window(
        self,
        start: pd.Timestamp,
        end: pd.Timestamp,
    ):
        """
        Extract indices of data points within a given time window.
        """
        mask = (self.corrected_neutron_data_frame.index >= start) & (
            self.corrected_neutron_data_frame.index <= end
        )
        return self.corrected_neutron_data_frame.index[mask].tolist()

    def extract_calibration_day_indices(self):
        """
        Extract indices for each calibration day within a 6-hour window.
        """
        unique_days = self._convert_to_datetime(
            self.calibration_data_prepper.unique_calibration_days
        )

        calibration_indices = {}
        for day in unique_days:
            start, end = self._create_time_window(day)
            calibration_indices[day] = self._extract_indices_within_window(
                start, end
            )

        return calibration_indices


class CalibrationWeightsCalculator:
    def __init__(
        self,
        time_series_data_object: PrepareNeutronCorrectedData,
        calib_data_object: PrepareCalibrationData,
        config: CalibrationConfiguration,
    ):
        self.time_series_data_object = time_series_data_object
        self.calib_data_object = calib_data_object
        self.config = config

        self.calib_metrics_dict = {}

    def _get_time_series_data_for_day(
        self,
        day,
    ):
        return self.time_series_data_object.data_dict[day]

    @staticmethod
    def _initial_vol_sm_estimate(profiles: List):
        """
        Gets an initial equal average soil moisture estimate

        Parameters
        ----------
        profiles : List
            List of SampleProfiles

        Returns
        -------
        sm_estimate : float
            Estimate of field soil moisture
        """
        sm_total_vol_values = [
            np.array(profile.sm_total_vol).flatten()
            for profile in profiles
            if profile.sm_total_vol is not None
        ]
        flattened = np.concatenate(sm_total_vol_values)
        valid_values = flattened[~np.isnan(flattened)]
        sm_estimate = np.mean(valid_values)
        return sm_estimate

    def apply_weighting_to_multiple_days(self):
        """
        Applies the weighting procedure to multiple calibration days if
        present.
        """

        for day in self.calib_data_object.unique_calibration_days:

            tmp_data = self._get_time_series_data_for_day(day)
            profiles = [
                p
                for p in self.calib_data_object.list_of_profiles
                if p.calibration_day == day
            ]

            volumetric_sm_estimate = self._initial_vol_sm_estimate(
                profiles=profiles
            )
            average_abs_air_humidity = tmp_data[
                self.config.abs_air_humidity_column
            ].mean()
            average_air_pressure = tmp_data[
                self.config.air_pressure_column_name
            ].mean()

            field_average_sm_vol, field_average_sm_grav, footprint = (
                self.calculate_weighted_sm_average(
                    profiles=profiles,
                    initial_volumetric_sm_estimate=volumetric_sm_estimate,
                    average_abs_air_humidity=average_abs_air_humidity,
                    average_air_pressure=average_air_pressure,
                )
            )

            output = {
                "field_average_soil_moisture_volumetric": field_average_sm_vol,
                "field_average_soil_moisture_gravimetric": field_average_sm_grav,
                "horizontal_footprint_radius_in_meters": footprint,
                "absolute_air_humidity": average_abs_air_humidity,
                "atmospheric_pressure": average_air_pressure,
            }

            self.calib_metrics_dict[day] = output

    def calculate_weighted_sm_average(
        self,
        profiles: List,
        initial_volumetric_sm_estimate: float,
        average_abs_air_humidity: float,
        average_air_pressure: float,
    ):
        """
        Calculates the field average soil moisture value according to
        Schrön et al., 2017. Weighting is calculated on volumetric soil
        moisture volumes (as described in the paper), but a weighted
        gravimetric value is given for calibration in neptoon.

        Parameters
        ----------
        profiles : List[Profile]
            A list of soil‐profile objects collected on the same day.
            Each Profile must have:
              - `.rescaled_distance` (rescaled distance from sensor, in
                m)
              - `.site_avg_bulk_density` (bulk density, in g/cm3)
              - `.depth` (array of depths, in cm)
              - `.sm_total_vol` (array of volumetric‐moisture values,
                cm3/cm3)
              - `.sm_total_grv` (array of gravimetric‐moisture values,
                g/g)
        initial_sm_estimate : float
            Initial soil moisture estimate (usually equal average)
        average_abs_air_humidity : float
            Average absolute air humidity
        average_air_pressure : float
            Air pressure average during calibration period (hPa)

        Returns
        -------
        field_average_sm_volumetric : float
            Converged volumetric soil moisture (cm3/cm3).
        field_average_sm_gravimetric : float
            Corresponding converged gravimetric soil moisture (g/g).
        footprint_m : float
            Estimated radius (m) of the footprint of the sensor.

        Notes
        -----
        - Convergence is checked via `abs((new_volumetric_estimate -
          old_estimate) / old_estimate) <
          self.config.converge_accuracy`.
        - Uses `Schroen2017.rescale_distance`,
          `Schroen2017.calculate_measurement_depth`,
          `Schroen2017.vertical_weighting`,
          `Schroen2017.horizontal_weighting`, and
          `Schroen2017.calculate_footprint_radius` at each iteration.

        """

        volumetric_sm_estimate = copy.deepcopy(initial_volumetric_sm_estimate)
        accuracy = 1
        field_average_sm_volumetric = None
        field_average_sm_gravimetric = None

        while accuracy > self.config.converge_accuracy:
            profile_sm_averages_volumetric = []
            profile_sm_averages_gravimetric = []
            profiles_horizontal_weights = []

            for p in profiles:

                p.rescaled_distance = Schroen2017.rescale_distance(
                    distance_from_sensor=p.rescaled_distance,
                    atmospheric_pressure=average_air_pressure,
                    volumetric_soil_moisture=volumetric_sm_estimate,
                )

                p.D86 = Schroen2017.calculate_measurement_depth(
                    distance=p.rescaled_distance,
                    bulk_density=p.site_avg_bulk_density,
                    volumetric_soil_moisture=volumetric_sm_estimate,
                )

                p.vertical_weights = Schroen2017.vertical_weighting(
                    depth=p.depth,
                    distance=p.rescaled_distance,
                    bulk_density=p.site_avg_bulk_density,
                    volumetric_soil_moisture=volumetric_sm_estimate,
                )

                # Calculate weighted sm average
                p.sm_total_weighted_avg_vol = np.average(
                    p.sm_total_vol, weights=p.vertical_weights
                )
                p.sm_total_weighted_avg_grv = np.average(
                    p.sm_total_grv, weights=p.vertical_weights
                )

                p.horizontal_weight = Schroen2017.horizontal_weighting(
                    distance=p.rescaled_distance,
                    volumetric_soil_moisture=p.sm_total_weighted_avg_vol,
                    abs_air_humidity=average_abs_air_humidity,
                )

                # create a list of average sm and horizontal weights
                profile_sm_averages_volumetric.append(
                    p.sm_total_weighted_avg_vol
                )
                profile_sm_averages_gravimetric.append(
                    p.sm_total_weighted_avg_grv
                )
                profiles_horizontal_weights.append(p.horizontal_weight)

            # mask out nan values from list
            profile_sm_averages_volumetric = np.ma.MaskedArray(
                profile_sm_averages_volumetric,
                mask=np.isnan(profile_sm_averages_volumetric),
            )
            profile_sm_averages_gravimetric = np.ma.MaskedArray(
                profile_sm_averages_gravimetric,
                mask=np.isnan(profile_sm_averages_gravimetric),
            )
            profiles_horizontal_weights = np.ma.MaskedArray(
                profiles_horizontal_weights,
                mask=np.isnan(profiles_horizontal_weights),
            )

            # create field averages of soil moisture

            field_average_sm_volumetric = np.average(
                profile_sm_averages_volumetric,
                weights=profiles_horizontal_weights,
            )
            field_average_sm_gravimetric = np.average(
                profile_sm_averages_gravimetric,
                weights=profiles_horizontal_weights,
            )

            # check convergence accuracy
            accuracy = abs(
                (field_average_sm_volumetric - volumetric_sm_estimate)
                / volumetric_sm_estimate
            )
            if accuracy > self.config.converge_accuracy:

                volumetric_sm_estimate = copy.deepcopy(
                    field_average_sm_volumetric
                )
                profile_sm_averages_volumetric = []
                profile_sm_averages_gravimetric = []
                profiles_horizontal_weights = []

        footprint_m = Schroen2017.calculate_footprint_radius(
            volumetric_soil_moisture=field_average_sm_volumetric,
            abs_air_humidity=average_abs_air_humidity,
            atmospheric_pressure=average_air_pressure,
        )

        return (
            field_average_sm_volumetric,
            field_average_sm_gravimetric,
            footprint_m,
        )

    def return_output_dict_as_dataframe(self):
        """
        Returns the dictionary of information created for each
        calibration day during processing as a pandas DataFrame

        Returns
        -------
        pd.DataFrame
            DataFrame with information created during processing.
        """
        df = pd.DataFrame.from_dict(self.calib_metrics_dict, orient="index")
        df = df.reset_index()
        df = df.rename(
            columns={
                "index": "calibration_day",
                "field_average_soil_moisture_volumetric": "field_average_soil_moisture_volumetric",
                "field_average_soil_moisture_gravimetric": "field_average_soil_moisture_gravimetric",
                "horizontal_footprint_in_meters": "horizontal_footprint_radius",
            }
        )
        return df

    def _find_optimal_N0_single_day_desilets_etal_2010(
        self,
        gravimetric_sm_on_day,
        neutron_mean,
        lattice_water=0,
        water_equiv_soil_organic_carbon=0,
    ):
        """
        Finds optimal N0 number when using desilets et al., 2010 method

        Parameters
        ----------
        gravimetric_sm_on_day : float
            gravimetric soil moisture (weighted)
        neutron_mean : float
            average (corrected) neutron count

        Returns
        -------
        float
            N0
        """

        n0_range = pd.Series(range(int(neutron_mean), int(neutron_mean * 2.5)))
        gravimetric_sm_on_day_total = (
            gravimetric_sm_on_day
            + lattice_water
            + water_equiv_soil_organic_carbon
        )

        def calculate_sm_and_error(n0):
            sm_prediction = (
                neutrons_to_total_grav_soil_moisture_desilets_etal_2010(
                    neutrons=neutron_mean,
                    n0=n0,
                )
            )
            absolute_error = abs(sm_prediction - gravimetric_sm_on_day_total)
            return pd.Series(
                {
                    "N0": n0,
                    "soil_moisture_prediction": sm_prediction,
                    "absolute_error": absolute_error,
                }
            )

        results_df = n0_range.apply(calculate_sm_and_error)
        min_error_idx = results_df["absolute_error"].idxmin()
        n0_optimal = results_df.loc[min_error_idx, "N0"]
        minimum_error = results_df.loc[min_error_idx, "absolute_error"]
        return n0_optimal, minimum_error

    def _find_optimal_n0_single_day_koehli_etal_2021(
        self,
        gravimetric_sm_on_day: float,
        neutron_mean: int,
        abs_air_humidity: float,
        lattice_water: float,
        water_equiv_soil_organic_carbon: float,
        koehli_method_form: str,
    ):
        """
        Finds optimal N0 number when using Koehli etal method

        Parameters
        ----------
        gravimetric_sm_on_day : float
            Average gravimetic water on calibration day)
        neutron_mean : float | int
            Mean corrected neutron count
        abs_air_humidity : float
            Absolute air humidity
        lattice_water : float
            Lattice water content of soil
        water_equiv_soil_organic_carbon : float
            water equivelant of soil organic carbon
        koehli_method_form: str
            The specific method form of Koehli method

        Returns
        -------
        Tuple
            The N0 calibration term and absolute error (dummy nan value)
        """
        n0 = compute_n0_koehli_etal_2021(
            gravimetric_sm=gravimetric_sm_on_day,
            neutron_count=neutron_mean,
            air_humidity=abs_air_humidity,
            lattice_water=lattice_water,
            water_equiv_soil_organic_carbon=water_equiv_soil_organic_carbon,
            koehli_method_form=koehli_method_form,
        )
        return n0, "nan"

    def find_optimal_N0(
        self,
    ):
        """
        Finds the optimal N0 number for the site using the weighted
        field average soil mositure.

        Returns
        -------
        average_n0
            The optimal n0 across all the supplied calibration days.
        """

        for day, metrics in self.calib_metrics_dict.items():
            neutron_mean = self.time_series_data_object.data_dict[day][
                self.config.neutron_column_name
            ].mean()
            grav_sm = metrics["field_average_soil_moisture_gravimetric"]

            if self.config.neutron_conversion_method == "desilets_etal_2010":
                n0_opt, abs_error = (
                    self._find_optimal_N0_single_day_desilets_etal_2010(
                        gravimetric_sm_on_day=grav_sm,
                        neutron_mean=neutron_mean,
                        lattice_water=self.config.value_avg_lattice_water,
                        water_equiv_soil_organic_carbon=self.config.value_avg_soil_organic_carbon_water_equiv,
                    )
                )

            elif self.config.neutron_conversion_method == "koehli_etal_2021":
                n0_opt, abs_error = (
                    self._find_optimal_n0_single_day_koehli_etal_2021(
                        gravimetric_sm_on_day=grav_sm,
                        neutron_mean=neutron_mean,
                        abs_air_humidity=metrics["absolute_air_humidity"],
                        lattice_water=self.config.value_avg_lattice_water,
                        water_equiv_soil_organic_carbon=self.config.value_avg_soil_organic_carbon_water_equiv,
                        koehli_method_form=self.config.koehli_method_form,
                    )
                )

            metrics.update({"optimal_N0": n0_opt, "absolute_error": abs_error})

        return float(
            np.mean(
                [m["optimal_N0"] for m in self.calib_metrics_dict.values()]
            )
        )



import pandas as pd
from pathlib import Path
import pytest
from neptoon.io.read.data_ingest import (
    TimeStampAligner,
)


def test_create_aligner_object_correct_format():
    """
    Test the creation of the object. Test checks for when data is
    formatted correctly.
    """
    data_path = Path(__file__).parent.parent / "test_data" / "io"
    data_correct_format = pd.read_csv(
        data_path / "unprocessed_df.csv",
        index_col=0,
        parse_dates=True,
    )
    tsa = TimeStampAligner(data_correct_format)
    assert isinstance(tsa, TimeStampAligner)


def test_create_aligner_object_wrong_format():
    """
    Test the creation of the object. Test checks for when data is not
    formatted correctly (i.e., not datetime index).
    """
    data_path = Path(__file__).parent.parent / "test_data" / "io"
    data_incorrect_format = pd.read_csv(
        data_path / "unprocessed_df.csv",
    )
    with pytest.raises(
        ValueError, match="The DataFrame index must be of datetime type"
    ):
        TimeStampAligner(data_incorrect_format)


# def test_align_timestamps():
#     """_summary_"""
#     data_path = Path(__file__).parent / "mock_data"
#     data_before_alignment = pd.read_csv(
#         data_path / "unprocessed_df.csv",
#         index_col=0,
#         parse_dates=True,
#     )
#     data_aligned = pd.read_csv(
#         data_path / "processed_df.csv",
#         index_col=0,
#         parse_dates=True,
#     )
#     tsa = TimeStampAligner(data_before_alignment)
#     tsa.align_timestamps()
#     result_df = tsa.return_dataframe()

#     pd.testing.assert_frame_equal(result_df, data_aligned, check_freq=False)


# test_align_timestamps()


def test_return_dataframe():
    data_path = Path(__file__).parent.parent / "test_data" / "io"
    data_correct_format = pd.read_csv(
        data_path / "unprocessed_df.csv",
        index_col=0,
        parse_dates=True,
    )
    tsa = TimeStampAligner(data_correct_format)
    df = tsa.return_dataframe()
    assert isinstance(df, pd.DataFrame)



import streamlit as st
from pathlib import Path

st.title(":material/save: Export")


if not st.session_state["data_corrections_made"]:
    st.warning("You need to process all neutron corrections first.")

if st.session_state["data_converted"]:

    @st.cache_data(show_spinner="Making figures...")
    def make_figures():
        st.session_state["yaml"].data_hub.create_figures(create_all=True)

    @st.cache_data(show_spinner="Save data...")
    def save_data():
        st.session_state["yaml"].data_hub.save_data()

    c11, c12, c13, c14 = st.columns(4)
    if c11.button("Make figures", type="primary"):
        make_figures()

    c21, c22, c23, c24 = st.columns(4)
    if c21.button("Save data", type="primary"):
        save_data()

    import glob

    sensor_name = st.session_state["yaml"].sensor_config.sensor_info.name
    matching_folders = glob.glob("{:}_*".format(sensor_name))
    if len(matching_folders) > 0:
        matching_numbers = [
            int(x[x.index("_") + 1 :]) for x in matching_folders
        ]
        latest_folder = "{:}_{:}".format(sensor_name, matching_numbers[-1])
        # st.write(latest_folder)

        pdf_report = "{:}/Report-{:}.pdf".format(latest_folder, sensor_name)
        if Path(pdf_report).is_file():
            with open(pdf_report, "rb") as file:
                btn = c12.download_button(
                    label=":material/description: PDF Report",
                    data=file,
                    file_name="Report-{:}.pdf".format(sensor_name),
                    mime="application/pdf",
                )

        csv_calibration = "{:}/data/{:}_calibration.csv".format(
            latest_folder, sensor_name
        )
        if Path(csv_calibration).is_file():
            with open(pdf_report, "rb") as file:
                btn = c22.download_button(
                    label=":material/table: Calibration data",
                    data=file,
                    file_name="{:}_calibration.csv".format(sensor_name),
                    mime="text/csv",
                )

        csv_flags = "{:}/data/{:}_flags.csv".format(latest_folder, sensor_name)
        if Path(csv_flags).is_file():
            with open(csv_flags, "rb") as file:
                btn = c23.download_button(
                    label=":material/table: Flag data",
                    data=file,
                    file_name="{:}_flags.csv".format(sensor_name),
                    mime="text/csv",
                )

        csv_processed_data = "{:}/data/{:}_processed_data.csv".format(
            latest_folder, sensor_name
        )
        if Path(csv_processed_data).is_file():
            with open(csv_processed_data, "rb") as file:
                btn = c24.download_button(
                    label=":material/table: Processed data",
                    data=file,
                    file_name="{:}_processed_data.csv".format(sensor_name),
                    mime="text/csv",
                )

        import os

        st.subheader("Image files")

        # Define the folder path
        image_folder = "{:}/figures/".format(latest_folder)

        # Get a list of image files
        image_files = [
            f for f in os.listdir(image_folder) if f.endswith(".png")
        ]

        for image in image_files:
            st.image(
                os.path.join(image_folder, image),
                caption=image,
                use_container_width=True,
            )



from __future__ import annotations

import os

from hatchling.version.source.plugin.interface import VersionSourceInterface


class EnvSource(VersionSourceInterface):
    PLUGIN_NAME = 'env'

    def get_version_data(self) -> dict:
        variable = self.config.get('variable', '')
        if not variable:
            message = 'option `variable` must be specified'
            raise ValueError(message)

        if not isinstance(variable, str):
            message = 'option `variable` must be a string'
            raise TypeError(message)

        if variable not in os.environ:
            message = f'environment variable `{variable}` is not set'
            raise RuntimeError(message)

        return {'version': os.environ[variable]}

    def set_version(self, version: str, version_data: dict) -> None:  # noqa: ARG002, PLR6301
        message = 'Cannot set environment variables'
        raise NotImplementedError(message)
